/*	Winnov Videum
 *	Video for Linux Two driver
 *
 *	This program is copyright 1998, 1999
 *	by Bill Dirks
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Video capture module for Winnov Videum. This module depends
 *	on wnv.o.
 */

/*  Support RGB capture formats. There is significant CPU overhead,
 *  but it may be needed for application compatibility.
 *  When libv4l2 is ready, this will not be needed because libv4l2
 *  can do the RGB conversion automatically  */
#define CONFIG_WNV_RGB


#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/sched.h>
#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/interrupt.h>

#define PKMOD "wnvvid: "
#include "wnv.h"

/*  Video controls  */

static struct v4l2_queryctrl videum_control[MAXCONTROLS] =
{
 {V4L2_CID_BRIGHTNESS, "Brightness", 0, 15, 1, 8, V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_CONTRAST,   "Contrast",   0, 15, 1, 4, V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_SATURATION, "Saturation", 0, 15, 1, 4, V4L2_CTRL_TYPE_INTEGER},
 {V4L2_CID_HUE,        "Hue",        0, 15, 1, 8, V4L2_CTRL_TYPE_INTEGER},
};
#define VCTRL_BRIGHTNESS	0
#define VCTRL_CONTRAST		1
#define VCTRL_SATURATION	2
#define VCTRL_HUE		3
static int
find_vctrl(int id)
{
	int i;
	if (id == V4L2_CID_PRIVATE_BASE ||
	    id <  V4L2_CID_BASE ||
            id >  V4L2_CID_LASTP1)
		return -EDOM;
	for (i = MAXCONTROLS - 1; i >= 0; i--)
		if (videum_control[i].id == id)
			break;
	if (i < 0)
		i = -EINVAL;
	return i;
}


static int unit_video[WNV_NBOARDS] = { 0, 1, 2, 3, };
MODULE_PARM(unit_video, "1-"__MODULE_STRING(WNV_NBOARDS)"i");

/*  Some config/debug options  */
extern int	wnv_noirq; 	/* 1 = Don't use hardware interrupts */
static int	slowdma	= 1; /* Fudge DMA rate by slowdma*10% 
			        (1 means allow 10% DMA rate variation) */

static inline struct videum_device *
videum_device_from_file(struct file *file)
{
	return (struct videum_device *)v4l2_device_from_file(file);
}


static void
wavi_brightness(struct videum_device *dev, int x)
{
	int	t;

	t = wavi_readreg(dev, WAVI_VOF);
	t &= 0xFF0F;
	x = x & 0x000F;
	wavi_writereg(dev, WAVI_VOF, t | (x << 4));
}

static void
wavi_contrast(struct videum_device *dev, int x)
{
	int	t;

	t = wavi_readreg(dev, WAVI_VCT);
	t &= 0xFF0F;
	x = x & 0x000F;
	wavi_writereg(dev, WAVI_VCT, t | (x << 4));
}

static void
wavi_saturation(struct videum_device *dev, int x)
{
	int	t;

	t = wavi_readreg(dev, WAVI_VCT);
	t &= 0x00FF;
	x = x & 0x000F;
	wavi_writereg(dev, WAVI_VCT, t | (x << 8) | (x << 12));
}

static void
wavi_hue(struct videum_device *dev, int x)
{
	int	t;

	t = wavi_readreg(dev, WAVI_VOF);
	t &= 0x00FF;
	x = x & 0x000F;
	wavi_writereg(dev, WAVI_VOF, t | (x << 8) | (x << 12));
}

static void
wavi_tone_controls(struct videum_device *dev)
{
	int	*ctrl;
	ctrl = dev->source[dev->input].control;
	wavi_brightness(dev, ctrl[VCTRL_BRIGHTNESS]);
	wavi_contrast(dev, ctrl[VCTRL_CONTRAST]);
	wavi_saturation(dev, ctrl[VCTRL_SATURATION]);
	wavi_hue(dev, ctrl[VCTRL_HUE]);
}

static void
wavi_vhsd(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_VCT) & 0xFFF0;
	wavi_writereg(dev, WAVI_VCT, t | (x & 0x000F));
}

static void
wavi_filter(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_CTL) & ~WAVI_VFSFIELD;
	t = (t & ~WAVI_IMIBIT) | dev->wavi_IMIBIT;
	wavi_writereg(dev, WAVI_CTL, t | x);
}

static void
wavi_vdf(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_CTL) & ~WAVI_VDFFIELD;
	t = (t & ~WAVI_IMIBIT) | dev->wavi_IMIBIT;
	wavi_writereg(dev, WAVI_CTL, t | x);
}

static void
wavi_vsc(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_CTL) & ~WAVI_VSCBIT;
	t = (t & ~WAVI_IMIBIT) | dev->wavi_IMIBIT;
	wavi_writereg(dev, WAVI_CTL, (x) ? t | WAVI_VSCBIT : t);
}

static void
wavi_vss(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_CTL) & ~WAVI_VSSBIT;
	t = (t & ~WAVI_IMIBIT) | dev->wavi_IMIBIT;
	wavi_writereg(dev, WAVI_CTL, (x) ? t | WAVI_VSSBIT : t);
}

static int
wavi_vst(struct videum_device *dev)
{
	return (wavi_readreg(dev, WAVI_CTL) & WAVI_VSTBIT);
}

static void
wavi_video_sampler(struct videum_device *dev, int x)
{
	__u16	t = wavi_readreg(dev, WAVI_CTL) & ~WAVI_VSEBIT;
	t = (t & ~WAVI_IMIBIT) | dev->wavi_IMIBIT;
	wavi_writereg(dev, WAVI_CTL, (x) ? t | WAVI_VSEBIT : t);
}

/*  Do an RVS instruction */
static void
wavi_do_rvs(struct videum_device *dev)
{
static	__u8	uc_rvs[] = {UI_RVS, UI_HLT, UI_HLT, UI_HLT,};
	int	timeout;

	if (!wavi_vst(dev))
		return;// already reset
	wavi_video_sampler(dev, 0);
	wavi_block_write(dev, dev->sram_ucode, uc_rvs, sizeof(uc_rvs), 1);
	wavi_writereg(dev, WAVI_SPLPTR, dev->sram_ucode >> 4);
	wavi_video_sampler(dev, 1);
	timeout = 10;
	while (timeout-- && wavi_vst(dev))
		continue;
	wavi_video_sampler(dev, 0);
	dev->uc_loaded = UC_UNDEF;
}

static u32 marker[4] = { 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff };
static void
wavi_write_markers(struct videum_device *dev,
		   __u32	start,
		   int		count,
		   __u32	interval)
{
	int	i;

	for (i = 0; i < count; ++i)
	{
		wavi_block_write(dev, start - sizeof(marker),
				 (__u8 *)marker, sizeof(marker), 1);
		start += interval;
	}
}

static int
isqrt(unsigned int q)
{/*	A little integer square root routine */
	int		i;
	int		r;
	unsigned int	b2	= 0x40000000;
	unsigned int	t;

	for (i = 16, r = 0; i > 0 && q; --i)
	{
		t = ((unsigned int)r << i) + b2;
		if (t <= q)
		{
			q -= t;
			r |= (1 << (i - 1));
		}
		b2 >>= 2;
	}
	return r;
}

static unsigned long
current_time_ms(void)
{
	struct timeval	now;

	do_gettimeofday(&now);
	return now.tv_sec * 1000 + now.tv_usec / 1000;
}


static void
rebuild_mmap_scatter_lists(struct videum_device *dev)
{
	/*  Rebuild the DMA scatter lists for the mmap'ed buffers
	    after the format has changed (capture_size has changed) */
	int			i;
	struct stream_buffer	*buf;

	if (dev->dma_mode == DMAMODE_SLAVE)
		return;
	for (i = 0; i < MAX_CAPTURE_BUFFERS; ++i)
	{
		buf = &dev->stream_buf[i];
		if (!buf->requested ||
		    !(buf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED) ||
		    buf->vaddress == NULL)
			continue;/* not mapped */

		wnv_bm_build_scatter_list(dev, buf->vaddress, &buf->dma_list);
	}
}


/*
 *
 *	V I D E O   D E C O D E R S
 *
 */

struct videc_reg_init
{
	__u16	reg;
	__u16	value;
};

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *	Philips 7110 Video Decoder routines
 */

static struct videc_reg_init
init_table_p7110[] =
{
	{0x00,	0x4C},		{0x01,	0x3C},
	{0x02,	0x0D},		{0x03,	0xEF},
	{0x04,	0xBD},		{0x05,	0xF0},
	{0x06,	0x00},		{0x07,	0x00},
	{0x08,	0xF8},		{0x09,	0xF8},
	{0x0A,	0x60},		{0x0B,	0x60},
	{0x0C,	0x00},		{0x0D,	0x86},
	{0x0E,	0x18},		{0x0F,	0x90},
	{0x10,	0x00},		{0x11,	0x2C},
	{0x12,	0x40},		{0x13,	0x46},
	{0x14,	0x42},		{0x15,	0x1A},
	{0x16,	0xFF},		{0x17,	0xDA},
	{0x18,	0xF0},		{0x19,	0x8B},
	{0x1A,	0x00},		{0x1B,	0x00},
	{0x1C,	0x00},		{0x1D,	0x00},
	{0x1E,	0x00},		{0x1F,	0x00},
	{0x20,	0xD9},		{0x21,	0x17},
	{0x22,	0x40},		{0x23,	0x41},
	{0x24,	0x80},		{0x25,	0x41},
	{0x26,	0x80},		{0x27,	0x4F},
	{0x28,	0xFE},		{0x29,	0x01},
	{0x2A,	0xCF},		{0x2B,	0x0F},
	{0x2C,	0x03},		{0x2D,	0x01},
	{0x2E,	0x81},		{0x2F,	0x03},
	{0x30,	0x44},		{0x31,	0x75},
	{0x32,	0x01},		{0x33,	0x8C},
	{0x34,	0x03},
};
#define NUM_P7110_REGISTERS	\
        (sizeof(init_table_p7110)/sizeof(struct videc_reg_init))

static int
p7110_initialize(struct videum_device *dev)
{
	int	i;

	/*  Video decoder information fields	*/
	dev->videc.standards		= (1 << V4L2_STD_NTSC) |
					  (1 << V4L2_STD_PAL);
	dev->videc.ntsc_hskip		= 30;
	dev->videc.ntsc_vskip		= 12;
	dev->videc.ntsc_width		= 640;
	dev->videc.ntsc_height		= 240;
	dev->videc.ntsc_field_order	= 0;
	dev->videc.pal_hskip		= 62;
	dev->videc.pal_vskip		= 14;
	dev->videc.pal_width		= 768;
	dev->videc.pal_height		= 288;
	dev->videc.pal_field_order	= 0;
	dev->videc.preferred_field	= 0;
	dev->videc.hadjust		= 0;
	dev->videc.vadjust		= 0;

	dev->videc.num_inputs		= 3;

	for (i = 0; i < VSOURCE_COUNT; ++i)
	{
		dev->source[i].wavi_vsc	= 0;
		dev->source[i].wavi_vss	= 0;
		dev->source[i].wavi_vhsd = 6;
		dev->source[i].vcrmode = 1;
	}

	for (i = 0; i < NUM_P7110_REGISTERS; ++i)
	{
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1,
				       init_table_p7110[i].reg,
				       init_table_p7110[i].value);
	}

	dev->videc.decoder_is_stable	= 1;

	return 1;
}

static int
p7110_set_input(struct videum_device *dev, int i)
{
	dev->input = i;
	switch (i)
	{
	case VSOURCE_MXC:
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x06, 0x00);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x20, 0xBA);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x21, 0x07);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x22, 0x91);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x2c, 0x03);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x30, 0x60);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x31, 0xB5);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x21, 0x05);
		break;
	case VSOURCE_COMP:
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x06, 0x00);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x20, 0xD9);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x21, 0x17);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x22, 0x40);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x2c, 0x03);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x30, 0x60);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x31, 0x75);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x21, 0x16);
		break;
	case VSOURCE_SVIDEO:
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x06, 0x80);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x20, 0x98);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x21, 0x17);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x22, 0x41);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x2c, 0x23);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x30, 0x44);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x31, 0x75);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x21, 0x14);
		break;
	}
	//dev->videc.decoder_is_stable = 0;
	return 1;
}

static int
p7110_set_standard(struct videum_device *dev, int x)
{
	switch (x)
	{
	case V4L2_STD_NTSC:
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x0D, 0x86);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x11, 0x2C);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x2E, 0x81);
		dev->videc.frame_period = 333667;
		break;
	case V4L2_STD_PAL:
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x0D, 0x86);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x11, 0x59);
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x2E, 0x9A);
		dev->videc.frame_period = 400000;
		break;
	case V4L2_STD_SECAM:
		dev->videc.frame_period = 400000;
		break;
	}
	dev->videc.standard = x;
	return 1;
}

static int
p7110_set_vcrmode(struct videum_device *dev, int x)
{
	/*  Leave 7110 always in VCR mode (7110 bug)	*/
	dev->source[dev->input].vcrmode = x;
	return 1;
}

static int
p7110_is_stable(struct videum_device *dev)
{
	return 1;
}

static int
p7110_probe(struct videum_device *dev)
{
	int	x;

	if (dev->videc.type == VIDEC_7110)
		return 1;/*  This device was already found  */
	if (dev->videc.type != VIDEC_BOGUS)
		return 0;/*  Another device was already found  */

	x = wnv_i2c_read_reg_byte(dev, 0x9C, 0);/*  Version  */
	if (x != -1)
	{
		dev->videc.i2c_addr1 = 0x9C;
	}
	else
	{
		x = wnv_i2c_read_reg_byte(dev, 0x9E, 0);
		if (x == -1)
			return 0;/*  Not found  */
		dev->videc.i2c_addr1 = 0x9E;
	}

	dev->videc.type = VIDEC_7110;

	/*  Fill in the method fields  */
	dev->videc.initialize = p7110_initialize;
	dev->videc.set_input = p7110_set_input;
	dev->videc.set_standard = p7110_set_standard;
	dev->videc.set_vcrmode = p7110_set_vcrmode;
	dev->videc.is_stable = p7110_is_stable;

	info_msg("Found Philips SAA7110 decoder chip at %2X\n",
		 dev->videc.i2c_addr1);
	return 1;/*  Found  */
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *	Philips 7111A Video Decoder routines
 */

static struct videc_reg_init
init_table_p7111a[] =
{
	{0x02,	0xC0},		{0x03,	0x23},
	{0x04,	0x00},		{0x05,	0x00},
	{0x06,	0xEB},		{0x07,	0xE0},
	{0x08,	0x88},		{0x09,	0x01},
	{0x0A,	0x80},		{0x0B,	0x47},
	{0x0C,	0x40},		{0x0D,	0x00},
	{0x0E,	0x01},
	{0x10,	0x40},		{0x11,	0x1C},
	{0x12,	0x01},		{0x13,	0x00},
	{0x14,	0x00},		{0x15,	0x00},
	{0x16,	0x00},		{0x17,	0x00},
};
#define NUM_P7111A_REGISTERS	\
        (sizeof(init_table_p7111a)/sizeof(struct videc_reg_init))

static int
p7111a_initialize(struct videum_device *dev)
{
	int	i;

	/*  Video decoder information fields	*/
	dev->videc.standards		= (1 << V4L2_STD_NTSC) |
					  (1 << V4L2_STD_PAL);
	dev->videc.ntsc_hskip		= 30;//72
	dev->videc.ntsc_vskip		= 10;//13
	dev->videc.ntsc_width		= 720;
	dev->videc.ntsc_height		= 240;
	dev->videc.ntsc_field_order	= 0;
	dev->videc.pal_hskip		= 30;
	dev->videc.pal_vskip		= 15;
	dev->videc.pal_width		= 720;
	dev->videc.pal_height		= 288;
	dev->videc.pal_field_order	= 0;
	dev->videc.preferred_field	= 0;
	dev->videc.hadjust		= 0;
	dev->videc.vadjust		= 0;

	dev->videc.num_inputs		= 3;

	for (i = 0; i < VSOURCE_COUNT; ++i)
	{
		dev->source[i].wavi_vsc	= 0;
		dev->source[i].wavi_vss	= 0;
		dev->source[i].wavi_vhsd = 6;
		dev->source[i].vcrmode = 0;
	}

	for (i = 0; i < NUM_P7111A_REGISTERS; ++i)
	{
		wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1,
				       init_table_p7111a[i].reg,
				       init_table_p7111a[i].value);
	}

	dev->videc.decoder_is_stable	= 1;

	return 1;
}

static int
p7111a_set_input(struct videum_device *dev, int i)
{
	int	m, s;

	dev->input = i;
	m = wnv_i2c_read_reg_byte(dev, dev->videc.i2c_addr1, 0x02) & 0xF8;
	s = wnv_i2c_read_reg_byte(dev, dev->videc.i2c_addr1, 0x09) & 0x7F;
	switch (i)
	{
	case VSOURCE_SVIDEO:
		m |= 0x05;
		s |= 0x80;
		break;
	case VSOURCE_COMP:
		m |= 0x00;
		s |= 0x00;
		break;
	case VSOURCE_MXC:
		m |= 0x02;
		s |= 0x00;
		break;
	}
	wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x02, m);
	wnv_i2c_write_reg_byte(dev, dev->videc.i2c_addr1, 0x09, s);
	//dev->videc.decoder_is_stable = 0;
	return 1;
}

static int
p7111a_set_standard(struct videum_device *dev, int x)
{
	/*  The initialize sets up the 7111A to detect the standard 
	    automatically, no need to do anything here on the 7111A */
	dev->videc.standard = x;
	switch (x)
	{
	case V4L2_STD_NTSC:
		dev->videc.frame_period = 333667;
		break;
	case V4L2_STD_PAL:
		dev->videc.frame_period = 400000;
		break;
	case V4L2_STD_SECAM:
		dev->videc.frame_period = 400000;
		break;
	}
	return 1;
}

static int
p7111a_set_vcrmode(struct videum_device *dev, int x)
{
	/*  Leave 7111A always in VCR mode	*/
	dev->source[dev->input].vcrmode = x;
	return 1;
}

static int
p7111a_is_stable(struct videum_device *dev)
{
	return 1;
}

static int
p7111a_probe(struct videum_device *dev)
{
	int	x;

	if (dev->videc.type == VIDEC_7111A)
		return 1;/*  This device was already found  */
	if (dev->videc.type != VIDEC_BOGUS)
		return 0;/*  Another device was already found  */

	x = wnv_i2c_read_reg_byte(dev, 0x48, 0);/*  Version  */
	if (x != -1)
	{
		dev->videc.i2c_addr1 = 0x48;
	}
	else
	{
		x = wnv_i2c_read_reg_byte(dev, 0x4A, 0);
		if (x == -1)
			return 0;/*  Not found  */
		dev->videc.i2c_addr1 = 0x4A;
	}

	dev->videc.type = VIDEC_7111A;

	/*  Fill in the method fields  */
	dev->videc.initialize = p7111a_initialize;
	dev->videc.set_input = p7111a_set_input;
	dev->videc.set_standard = p7111a_set_standard;
	dev->videc.set_vcrmode = p7111a_set_vcrmode;
	dev->videc.is_stable = p7111a_is_stable;

	info_msg("Found Philips SAA7111A decoder chip at %2X\n",
		 dev->videc.i2c_addr1);
	return 1;/*  Found  */
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *	ITT VPX Video Decoder routines
 */

static struct videc_reg_init
init_table_ittvpx[] =
{/*  Registers 0x1000-0x11FF are FP regs, 0x000-0x0FF I2C regs */
	{0x10D0,	0xFFF},
	{0x101C,	0x000},
	{0x10A0,	0x700},
	{0x1088,	0x00C},
	{0x1089,	0x0F8}, /* PAL: 128 */
	{0x108A,	0x0F8}, /* PAL: 128 */
	{0x108B,	0x000},
	{0x108C,	0x420},
	{0x108D,	0x420},
	{0x108F,	0x000},
	{0x10F0,	0x076},
	{0x10F2,	0x3D3}, /* PAL: 3DB */
	{0x10B2,	0x300},
	{0x0033,	0x001},
	{0x00D8,	0x00C},
	{0x00D0,	0x040},
	{0x00D1,	0x040},
	{0x00E6,	0x008},
	{0x00E7,	0x020},
	{0x00E8,	0x0F8},
	{0x00F0,	0x00B},
	{0x00F1,	0x018},
	{0x00F2,	0x01B},
	{0x00F8,	0x0FA},
};
#define NUM_ITTVPX_REGISTERS	\
        (sizeof(init_table_ittvpx)/sizeof(struct videc_reg_init))

static int
ittvpx_read_reg(struct videum_device *dev, int reg)
{
	int	addr	= dev->videc.i2c_addr1;
	if (reg & 0x1000)
	{
		int	x = wnv_i2c_read_reg_byte(dev, addr, 0x29);
		int	timeout;
		if (x < 0)
			return x;
		for (timeout = 200; timeout && (x & 7) != 0; --timeout)
			x = wnv_i2c_read_reg_byte(dev, addr, 0x29);
		reg &= ~0x1000;
		wnv_i2c_write_reg_word(dev, addr, 0x26, reg);
		return wnv_i2c_read_reg_word(dev, addr, 0x28);
	}
	return wnv_i2c_read_reg_byte(dev, addr, reg);
}

static int
ittvpx_write_reg(struct videum_device *dev,
		 int reg, int ndata)
{
	int	addr	= dev->videc.i2c_addr1;
	if (reg & 0x1000)
	{
		int	x = wnv_i2c_read_reg_byte(dev, addr, 0x29);
		int	timeout;
		if (x < 0)
			return x;
		for (timeout = 200; timeout && (x & 7) != 0; --timeout)
			x = wnv_i2c_read_reg_byte(dev, addr, 0x29);
		reg &= ~0x1000;
		wnv_i2c_write_reg_word(dev, addr, 0x27, reg);
		return wnv_i2c_write_reg_word(dev, addr, 0x28, ndata);
	}
	return wnv_i2c_write_reg_byte(dev, addr, reg, ndata);
}

static int
ittvpx_initialize(struct videum_device *dev)
{
	int	i;

	/*  Video decoder information fields	*/
	dev->videc.standards		= (1 << V4L2_STD_NTSC)  |
					  (1 << V4L2_STD_PAL)   |
					//(1 << V4L2_STD_SECAM) |
		                          0;
	dev->videc.ntsc_hskip		= 16;
	dev->videc.ntsc_vskip		= 14;
	dev->videc.ntsc_width		= 1025;
	dev->videc.ntsc_height		= 240;
	dev->videc.ntsc_field_order	= 0;
	dev->videc.pal_hskip		= 16;
	dev->videc.pal_vskip		= 14;
	dev->videc.pal_width		= 1025;
	dev->videc.pal_height		= 288;
	dev->videc.pal_field_order	= 0;
	dev->videc.preferred_field	= 0;
	dev->videc.hadjust		= 0;
	dev->videc.vadjust		= 0;

	dev->videc.num_inputs		= 3;

	for (i = 0; i < VSOURCE_COUNT; ++i)
	{
		dev->source[i].wavi_vsc	= 0;
		dev->source[i].wavi_vss	= 0;
		dev->source[i].wavi_vhsd = 0;
		dev->source[i].vcrmode = 1;
	}

	for (i = 0; i < NUM_ITTVPX_REGISTERS; ++i)
	{
		ittvpx_write_reg(dev,
				 init_table_ittvpx[i].reg,
				 init_table_ittvpx[i].value);
	}

	dev->videc.decoder_is_stable	= 1;

	return 1;
}

static int
ittvpx_set_input(struct videum_device *dev, int i)
{
	int	svhs = 0, luma = 1, chroma = 0;
	int	x;

	dev->input = i;
	switch (i)
	{
	case VSOURCE_MXC:    luma = 1;		 break;
	case VSOURCE_COMP:   luma = 3;		 break;
	case VSOURCE_SVIDEO: luma = 2; svhs = 1; break;
	}
	luma = 3 - luma;
	chroma = (1 - chroma) << 2;
	svhs <<= 5;

	/*  input multiplexers  */
	x = ittvpx_read_reg(dev, 0x33);
	ittvpx_write_reg(dev, 0x33, (x & 0xF8) | 0x08 | chroma | luma);
	/*  SVHS mode  */
	x = ittvpx_read_reg(dev, 0x10F2);
	ittvpx_write_reg(dev, 0x10F2, (x & ~0x20) | svhs | 0x10);

	//dev->videc.decoder_is_stable = 0;
	return 1;
}

static int
ittvpx_set_standard(struct videum_device *dev, int x)
{
	switch (x)
	{
	case V4L2_STD_NTSC:
		ittvpx_write_reg(dev, 0x1089, 0x0F8);
		ittvpx_write_reg(dev, 0x108A, 0x0F8);
		ittvpx_write_reg(dev, 0x10F2, 0x3D3);
		dev->videc.frame_period = 333667;
		break;
	case V4L2_STD_PAL:
		ittvpx_write_reg(dev, 0x1089, 0x128);
		ittvpx_write_reg(dev, 0x108A, 0x128);
		ittvpx_write_reg(dev, 0x10F2, 0x3DB);
		dev->videc.frame_period = 400000;
		break;
	case V4L2_STD_SECAM:/* BUG: this is not right for SECAM */
		ittvpx_write_reg(dev, 0x1089, 0x128);
		ittvpx_write_reg(dev, 0x108A, 0x128);
		ittvpx_write_reg(dev, 0x10F2, 0x3DB);
		dev->videc.frame_period = 400000;
		break;
	}
	dev->videc.standard = x;
	return 1;
}

static int
ittvpx_set_vcrmode(struct videum_device *dev, int x)
{
	dev->source[dev->input].vcrmode = x;
	ittvpx_write_reg(dev, 0x104B, (x) ? 0x2FF : 0x29C);
	return 1;
}

static int
ittvpx_is_stable(struct videum_device *dev)
{
	return 1;
}

static int
ittvpx_probe(struct videum_device *dev)
{
	int	x;

	if (dev->videc.type == VIDEC_VPX)
		return 1;/*  This device was already found  */
	if (dev->videc.type != VIDEC_BOGUS)
		return 0;/*  Another device was already found  */

	dev->videc.i2c_addr1 = 0x86;
	x = ittvpx_read_reg(dev, 0x1050);/*  Version  */
	if (x == -1)
	{
		dev->videc.i2c_addr1 = 0x8E;
		x = ittvpx_read_reg(dev, 0x1050);
		if (x == -1)
		{
			dev->videc.i2c_addr1 = 0;
			return 0;/*  Not found  */
		}
	}

	dev->videc.type = VIDEC_VPX;
	dev->videc.version = x;

	/*  Fill in the method fields  */
	dev->videc.initialize = ittvpx_initialize;
	dev->videc.set_input = ittvpx_set_input;
	dev->videc.set_standard = ittvpx_set_standard;
	dev->videc.set_vcrmode = ittvpx_set_vcrmode;
	dev->videc.is_stable = ittvpx_is_stable;

	info_msg("Found ITT VPX (ver.%d) decoder chip at %2X\n",
		 dev->videc.version, dev->videc.i2c_addr1);
	if (x == 700)
		err_msg("Old VPX version not correctly supported.\n");

	return 1;/*  Found  */
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *
 *	Probe I2C bus for video decoder and fill in the device fields
 */
static int
find_decoder(struct videum_device *dev)
{
	if (!p7110_probe(dev) &&
	    !p7111a_probe(dev) &&
	    !ittvpx_probe(dev))
		return 0;/*  Failure  */
	return 1;
}

static void
set_video_input(struct videum_device *dev,
		int i)
{
	if (i < 0 || i >= dev->videc.num_inputs)
		return;
	dev->videc.set_input(dev, i);
	/*dev->videc.set_vcrmode(dev, dev->source[i].vcrmode);*/
	wavi_vsc(dev, dev->source[i].wavi_vsc);
	wavi_vss(dev, dev->source[i].wavi_vss);
	wavi_vhsd(dev, dev->source[i].wavi_vhsd);
	wavi_tone_controls(dev);
}

/*
 *
 *	W A V I   V I D E O   C A P T U R E   M I C R O C O D E
 *
 */

/*  Video line capture modes	*/
#define VMODE_NORMAL	0  /* Capture incoming pixels to VID1PTR */
#define VMODE_AVGBUF	1  /* Capture incoming pixels to VID2PTR */
#define VMODE_AVG	2  /* Average incoming with VID2PTR to VID1PTR */
#define VMODE_SKIP	(~0)  /* Fake mode: Skip whole line */

struct uc_buffer
{
	__u8	*p;	/* Current position	*/
	int	len;	/* Bytes remaining	*/
};

static void
compile_repeat(struct uc_buffer *buf,
	       __u8	ui,
	       int	n)
{
	if (buf->len < n)
		n = buf->len;
	buf->len -= n;
	while (n--)
		*(buf->p)++ = ui;
}

static void
compile_start_of_field(struct uc_buffer	*buf,
		       int	nfield,
		       int	vskip)
{
	if (vskip < 0 || vskip > 256)
		vskip = 0;
	if (buf->len < 4 + vskip)
		return;

	*(buf->p)++ = UI_WTVB;
	if (nfield == 1)
		*(buf->p)++ = UI_WTVF1;
	else
		*(buf->p)++ = UI_WTVF2;
	*(buf->p)++ = UI_WTVB;
	*(buf->p)++ = UI_WTVA;
	compile_repeat(buf, UI_WTHS, vskip);
	buf->len -= 4 + vskip;
}

static void
compile_interrupt(struct uc_buffer *buf,
		  __u8	ui_interrupt,
		  int	flags)
{
	if (buf->len < 3)
		return;

	*(buf->p)++ = ui_interrupt;
	--buf->len;
	if (flags & COMPFLAG_AVPRO_INT)
	{
		*(buf->p)++ = UI_SVS;
		*(buf->p)++ = UI_RVS;
		buf->len -= 2;
	}
}

static void
compile_dma_trigger(struct uc_buffer *buf,
		    int		flags)
{
	if (buf->len < 3)
		return;

	if (flags & COMPFLAG_AVPRO_DMA)
	{
		*(buf->p)++ = UI_SVS;
		*(buf->p)++ = UI_SVS;
		*(buf->p)++ = UI_RVS;
		buf->len -= 3;
	}
}

static void
compile_set_videomode(struct uc_buffer *buf,
		      int	*videomode,
		      int	newmode)
{
	int	n = (newmode - *videomode) & 3;

	if (n == 0 || buf->len < n)
		return;

	buf->len -= n;
	while (n--)
		*(buf->p)++ = UI_SVM;
	*videomode = newmode;
}

static void
compile_xcrop(struct uc_buffer *buf,
	      int	x)
{
	if (x < 1 || buf->len < 2)
		return;

	if (x > 960)
		x = 960;
	*((__u16 *)buf->p)++ = UI_SKIP_LOWPREC(x);
	buf->len -= 2;
}

static void
compile_horizontal_decimation(
	int	source_width,
	int	capture_width,
	int	*hrate,
	int	*hclocks,
	int	flags)
{
	int	npx;

	if (source_width == 0)
		return;
	*hclocks = source_width;
	*hrate = (capture_width * 64) / source_width;
	if (*hrate < 2)
		*hrate = 2;
	npx = (*hclocks * *hrate + 32) / 64;
	/*  Force the decimation ratio up if not enough pixels captured	*/
	while (npx < capture_width)
	{
		(*hrate)++;
		npx = (*hclocks * *hrate + 32) / 64;
	}
	/*  Force the number of clocks down if too many pixels captured	*/
	while (npx > capture_width)
	{
		(*hclocks)--;
		npx = ((*hclocks * *hrate) + 32) / 64;
	}
	    
	/*  Attempt to work around WAVI video sampler bugs:	*/
	/*  Capture for the maximum number of clocks that take the number of */
	/*  pixels we want		*/
	while (((*hclocks + 1) * *hrate + 32) / 64 == npx)
		++(*hclocks);
}

static void
compile_line(struct uc_buffer *buf,
	     int	hskip,
	     __u16	ui_capture1,
	     __u16	ui_capture2,
	     int	*videomode,
	     int	vmode)
{
	int	x;

	if (buf->len < 15)/*  No room at the inn...  */
		return;

	if (vmode == VMODE_SKIP)/*  Don't capture any pixels on this line  */
	{
		*(buf->p)++ = UI_WTHS;
		--buf->len;
		return;
	}

	compile_set_videomode(buf, videomode, vmode);

	*(buf->p)++ = UI_WTHS;
	--buf->len;

	x = hskip;
	if (vmode == VMODE_AVGBUF || vmode == VMODE_AVG)
	{/*	Rewind VID2PTR before capturing this line	*/
		*(buf->p)++ = UI_RVP2;
		--buf->len;
		--x;
	}

	compile_xcrop(buf, x);

	*((__u16 *)buf->p)++ = ui_capture1;
	buf->len -= 2;
	if (ui_capture2)
	{
		*((__u16 *)buf->p)++ = ui_capture2;
		buf->len -= 2;
	}
}

static void
compile_field(struct uc_buffer *buf,
	      int	source_height,
	      int	capture_height,
	      int	hskip,
	      __u16	ui_capture1,
	      __u16	ui_capture2,
	      int	interrupt_line,
	      int	dma_line,
	      int	flags)
{
	int	a, m;
	int	source_line_count	= 0;
	int	captured_line_count	= 0;
	int	avg_buf_full		= 0;
	int	videomode		= 0;

	a = source_height / 2;
	while (captured_line_count < capture_height)
	{
		if (a >= source_height)
		{/*	Capture this line to output buffer	*/
			a -= source_height;
			++captured_line_count;
			m = (avg_buf_full) ? VMODE_AVG : VMODE_NORMAL;
			avg_buf_full = 0;
		}
		else
		{/*	Capture to average buffer or just skip	*/
			if (avg_buf_full || (flags & COMPFLAG_NOVERTFILTER))
				m = VMODE_SKIP;
			else
			{
				m = VMODE_AVGBUF;
				avg_buf_full = 1;
			}
		}

		compile_line(buf, hskip, ui_capture1, ui_capture2,
			     &videomode, m);

		++source_line_count;
		if (source_line_count == dma_line)
			compile_dma_trigger(buf, flags);
		if (source_line_count == interrupt_line)
			compile_interrupt(buf, UI_ITR1, flags);

		a += capture_height;
	}
	compile_set_videomode(buf, &videomode, 0);
	if ((flags & COMPFLAG_HUFFMAN) && buf->len > 1)
	{
		*((__u16 *)buf->p)++ = UI_DECIMATE_LOWPREC(32, 64);
		buf->len -= 2;
	}

	while ( source_line_count < interrupt_line ||
		source_line_count < dma_line)
	{
		compile_line(buf, 0, 0, 0, 0, VMODE_SKIP);
		++source_line_count;
		if (source_line_count == dma_line)
			compile_dma_trigger(buf, flags);
		if (source_line_count == interrupt_line)
			compile_interrupt(buf, UI_ITR1, flags);
	}
}

static int/*  Number of bytes of code generated  */
compile_microcode(struct videum_device *dev,
		  struct ucode_parms *parm)
{
	struct uc_buffer buffer;
	int	hrate;
	int	hclocks;
	int	field1		= 1;
	__u16	ui_capture1	= 0;
	__u16	ui_capture2	= 0;

	buffer.p = parm->buffer;
	buffer.len = dev->sram_ucode_size;

	if (parm->flags & COMPFLAG_VST_WHEN_DONE)
	{
		*(buffer.p)++ = UI_RVS;
		--buffer.len;
	}

	compile_horizontal_decimation(parm->source_width, parm->capture_width,
				      &hrate, &hclocks, parm->flags);
	if (!((parm->flags & COMPFLAG_HUFFMAN) && hrate > 32) ||
	     ((parm->flags & COMPFLAG_HUFFMAN) && hrate > 19))
		parm->flags |= COMPFLAG_NOVERTFILTER;
	if (hclocks > 960)
	{
		ui_capture1 = UI_DECIMATE_LOWPREC(hrate, hclocks/2);
		ui_capture2 = UI_DECIMATE_LOWPREC(hrate, hclocks - hclocks/2);
	}
	else
	{
		ui_capture1 = UI_DECIMATE_LOWPREC(hrate, hclocks);
	}
	/*  Center the actual sample range within desired sample range  */
	if (hclocks > parm->source_width)
		hclocks = parm->source_width;
	parm->source_x += (parm->source_width - hclocks) / 2;
	/*  Horizontal filter */
	parm->filter = 0;
	if (hrate < 48)
		parm->filter |= WAVI_VFS_MED3;
	if (hrate < 16)
		parm->filter |= WAVI_VFS_FIR4;
	else if (hrate < 32)
		parm->filter |= WAVI_VFS_FIR2;

	field1 = parm->field;
	compile_start_of_field(&buffer, field1, parm->source_y);
	compile_field(&buffer,  
		      parm->source_height, parm->capture_height,
		      parm->source_x, ui_capture1, ui_capture2,
		      parm->int1_line, parm->dma_line,
		      parm->flags);

	if (buffer.len < 16)
	{
		err_msg("Microcode buffer too small.\n");
		return 0;/*  Ran out of buffer! Oh, no!  */
	}

	if (parm->flags & COMPFLAG_VST_WHEN_DONE)
	{
		*(buffer.p)++ = UI_SVS;
		--buffer.len;
	}
	*(buffer.p)++ = UI_HLT;
	--buffer.len;
	return ((dev->sram_ucode_size - buffer.len) + 3) & ~3;
}


/*
 *
 *	I M A G E   F O R M A T   T R A N S L A T I O N
 *
 */

static int
translate_yuyv_grey(struct translation *xlat)
{
	__u8	*esi, *edi;
	__u32	eax, ebx, ecx, edx;
	int	row;

	esi = xlat->in;
	edi = xlat->out;
	eax = xlat->in_stride - xlat->width * 2;
	ebx = xlat->out_stride - xlat->width;
	for (row = xlat->height; row; esi += eax, edi += ebx, --row)
	{
		for (ecx = xlat->width >> 2; ecx; --ecx)
		{
			edx = esi[4] | (esi[6] << 8);
			edx <<= 16;
			edx |= esi[0] | (esi[2] << 8);
			esi += 8;
			*(__u32 *)edi = edx;
			edi += 4;
		}
	}
	return 1;
}

static int
translate_yuyv_yuv420(struct translation *xlat)
{
	__u8	*esi, *edi;
	__u32	eax, ebx, ecx, edx;
	__u8	dl;
	int	row;

	/* Y's */
	esi = xlat->in;
	edi = xlat->out;
	eax = xlat->in_stride - xlat->width * 2;
	ebx = xlat->out_stride - xlat->width;
	for (row = xlat->height; row; esi += eax, edi += ebx, --row)
	{
		for (ecx = xlat->width >> 2; ecx; --ecx)
		{
			edx = esi[4] | (esi[6] << 8);
			edx <<= 16;
			edx |= esi[0] | (esi[2] << 8);
			esi += 8;
			*(__u32 *)edi = edx;
			edi += 4;
		}
	}
	/* U's */
	esi = xlat->in + xlat->in_stride;
	eax = xlat->in_stride * 2 - xlat->width * 2;
	ebx >>= 1;
	for (row = xlat->height >> 1; row; esi += eax, edi += ebx, --row)
	{
		for (ecx = xlat->width >> 1; ecx; ++edi, --ecx)
		{
			dl = esi[1];
			esi += 4;
			*edi = dl;
		}
	}
	/* V's */
	esi = xlat->in + xlat->in_stride;
	for (row = xlat->height >> 1; row; esi += eax, edi += ebx, --row)
	{
		for (ecx = xlat->width >> 1; ecx; ++edi, --ecx)
		{
			dl = esi[3];
			esi += 4;
			*edi = dl;
		}
	}
	return 1;
}

#define K12_1	 4096
#define K12_S	   12
#define K12_GU	-1409
#define K12_BU	 7258
#define K12_RV	 5743
#define K12_GV	-2925
static int
translate_expand_y(int y)
{
	y = (255 * y + 110) / 220;
	if (y < 0) y = 0; else if (y > 255) y = 255;
	return y;
}
static int
translate_expand_c(int c)
{
	c = (127 * c + 56) / 112;
	if (c < -128) c = -128; else if (c > 127) c = 127;
	return c;
}

static int
translate_make_rgb16_lut(struct translation *xlat)
{
	__u16	*lut;
	long	gu[32], bu[32], rv[32], gv[32];
	int	rscale[256], gscale[256], bscale[256];
	int	rrange, grange, brange;
	int	rshift, gshift, bshift;
	long	x;
	int	y, u, v;
	int	r, g, b;
	int	i, t;

	if ((xlat->type == XLAT_YUYV_TO_RGB555 &&
	     xlat->lut[0].type == LUT_RGB555) ||
	    (xlat->type == XLAT_YUYV_TO_RGB565 &&
	     xlat->lut[0].type == LUT_RGB565))
		return 1;

	if (xlat->lut[0].table.base)
		vfree(xlat->lut[0].table.base);
	xlat->lut[0].table.base = vmalloc(1 << 17);
	if (xlat->lut[0].table.base == NULL)
	{
		err_msg("vmalloc() failed in make_rgb16_lut\n");
		return 0;
	}
	lut = xlat->lut[0].table.rgb16;

	//  Compute all different chroma components to 8-bit precision
	for (i = 0, t = -128; t < 128; t += 8, ++i)
	{
		x = translate_expand_c(t) + 2;
		gu[i] = (K12_GU * x + K12_1/2) >> K12_S;
		bu[i] = (K12_BU * x + K12_1/2) >> K12_S;
		rv[i] = (K12_RV * x + K12_1/2) >> K12_S;
		gv[i] = (K12_GV * x + K12_1/2) >> K12_S;
	}
	//  8-bit to ?-bit scaling tables
	if (xlat->type == XLAT_YUYV_TO_RGB555)
	{
		xlat->lut[0].type = LUT_RGB555;
		rrange = grange = brange = 31;
		rshift = 10; gshift = 5; bshift = 0;
	}
	else
	{
		xlat->lut[0].type = LUT_RGB565;
		rrange = brange = 31;
		grange = 63;
		rshift = 11; gshift = 5; bshift = 0;
	}
	for (i = 0; i < 256; ++i)
	{
		rscale[i] = ((i * rrange + 127) / 255) << rshift;
		gscale[i] = ((i * grange + 127) / 255) << gshift;
		bscale[i] = ((i * brange + 127) / 255) << bshift;
	}

	//  Fill in the RGB values for each combination of YUV
	for (i = 0; i < 256; i += 4)
	{
		y = translate_expand_y(i) + 2;
		if (y > 255) y = 255;
		for (u = 0; u < 32; ++u)
			for (v = 0; v < 32; ++v)
			{
				//  Red, Green and Blue
				r = y + rv[v];
				g = y + gu[u] + gv[v];
				b = y + bu[u];
				//  Saturate
				if (r < 0) r = 0; else if (r > 254) r = 254;
				if (g < 0) g = 0; else if (g > 254) g = 254;
				if (b < 0) b = 0; else if (b > 254) b = 254;
				//  scale, shift and combine
				*lut++ = rscale[r] + gscale[g] + bscale[b];
			}
	}
	return 1;
}

static int
translate_yuyv_rgb16(struct translation *xlat)
{
	__u32	*src, *dst;
	__u32	uv, yuv0, yuv1, dual;
	__u16	*lut;
	int	stride;
	int	row, i;

	if (!translate_make_rgb16_lut(xlat))
		return 0;

	src = (__u32 *)xlat->in;
	dst = (__u32 *)xlat->out;
	lut = xlat->lut[0].table.rgb16;
	stride = (xlat->out_stride - xlat->width * 2) >> 2;
	for (row = xlat->height; row; --row)
	{
		for (i = xlat->width >> 1; i; --i)
		{
			dual = *src++;
			uv   =    ((dual & 0x0000F800) >> 6)
				+ ((dual & 0xF8000000) >> 27);
			yuv1 = 	  ((dual & 0x00FC0000) >> 8) + uv;
			yuv0 = 	  ((dual & 0x000000FC) << 8) + uv;
			*dst++ = ((__u32)lut[yuv1] << 16) | lut[yuv0];
		}
		dst += stride;
	}
	return 1;
}

static int
translate_make_rgb24_lut(struct translation *xlat)
{
	struct lookup_rgb24	*lut;
	int			r, g, b;
	int			i;
	int			x;

	if (xlat->lut[0].type == LUT_RGB24)
		return 1;

	if (xlat->lut[0].table.base)
		vfree(xlat->lut[0].table.base);
	xlat->lut[0].table.base = vmalloc(sizeof(struct lookup_rgb24));
	if (xlat->lut[0].table.base == NULL)
	{
		err_msg("vmalloc() failed in make_rgb24_lut\n");
		return 0;
	}
	xlat->lut[0].type = LUT_RGB24;
	lut = xlat->lut[0].table.rgb24;

	for (i = 0; i < 256; ++i)
	{
		x = i;		// Value is in excess-128 format
		if (x < 128)
			++x;	// Add 1 to negative values for noise rejection
		x -= 128;	// Convert to two's complement format
		x = translate_expand_c(x);

		g = (K12_GU * x + K12_1/2) >> K12_S;
		b = (K12_BU * x + K12_1/2) >> K12_S;
		lut->u_rgb[i] = ((g & 0x3FF) << 11) | (b & 0x3FF);

		r = (K12_RV * x + K12_1/2) >> K12_S;
		g = (K12_GV * x + K12_1/2) >> K12_S;
		lut->v_rgb[i] = (r << 22) | ((g & 0x3FF) << 11);
                
		x = translate_expand_y(i);
		lut->y_rgb[i] = (x << 22) | (x << 11) | x;
	}
	for (i = 0; i < 1024; ++i)
	{
		x = (i > 511) ? 0 : ((i > 255) ? 255 : i);
		lut->sat[i] = x;
	}
	return 1;
}

static int
translate_yuyv_rgb24(struct translation *xlat)
{
	struct lookup_rgb24	*lut;
	__u8			*src;
	__u32			*dst;
	int			i;
	int			row;
	int			stride;
	__u32			pela, pelb, pelc, peld;

	if (!translate_make_rgb24_lut(xlat))
		return 0;
	lut = xlat->lut[0].table.rgb24;
	src = (__u8 *)xlat->in;
	dst = (__u32 *)xlat->out;
	stride = (xlat->out_stride - 3 * xlat->width) >> 2;
	for (row = xlat->height; row; --row)
	{
		for (i = xlat->width >> 2; i; --i)
		{
			pelb  = lut->u_rgb[src[1]]
			      + lut->v_rgb[src[3]];
			pela  = lut->y_rgb[src[0]] + pelb;
			pelb += lut->y_rgb[src[2]];
			peld  = lut->u_rgb[src[5]]
			      + lut->v_rgb[src[7]];
			pelc  = lut->y_rgb[src[4]] + peld;
			peld += lut->y_rgb[src[6]];
			src += 8;

			dst[0] = ((u32)lut->sat[pela & 0x3FF])
			       + ((u32)lut->sat[(pela >> 11) & 0x3FF] << 8)
			       + ((u32)lut->sat[pela >> 22] << 16)
			       + ((u32)lut->sat[pelb & 0x3FF] << 24);
			dst[1] = ((u32)lut->sat[(pelb >> 11) & 0x3FF])
			       + ((u32)lut->sat[pelb >> 22] << 8)
			       + ((u32)lut->sat[pelc & 0x3FF] << 16)
			       + ((u32)lut->sat[(pelc >> 11) & 0x3FF] << 24);
			dst[2] = ((u32)lut->sat[pelc >> 22])
			       + ((u32)lut->sat[peld & 0x3FF] << 8)
			       + ((u32)lut->sat[(peld >> 11) & 0x3FF] << 16)
			       + ((u32)lut->sat[peld >> 22] << 24);
			dst += 3;
		}
		dst += stride;
	}

	return 1;
}

static int
translate_yuyv_rgb32(struct translation *xlat)
{
	struct lookup_rgb24	*lut;
	__u8			*src;
	__u32			*dst;
	int			i;
	int			row;
	int			stride;
	__u32			pela, pelb;

	if (!translate_make_rgb24_lut(xlat))
		return 0;
	lut = xlat->lut[0].table.rgb24;
	src = (__u8 *)xlat->in;
	dst = (__u32 *)xlat->out;
	stride = (xlat->out_stride - 4 * xlat->width) >> 2;
	for (row = xlat->height; row; --row)
	{
		for (i = xlat->width >> 1; i; --i)
		{
			pelb  = lut->u_rgb[src[1]]
			      + lut->v_rgb[src[3]];
			pela  = lut->y_rgb[src[0]] + pelb;
			pelb += lut->y_rgb[src[2]];
			src += 4;
			
			dst[0] =  (u32)lut->sat[pela & 0x3FF]
			       + ((u32)lut->sat[(pela >> 11) & 0x3FF] << 8)
			       + ((u32)lut->sat[pela >> 22] << 16);
			dst[1] =  (u32)lut->sat[pelb & 0x3FF]
			       + ((u32)lut->sat[(pelb >> 11) & 0x3FF] << 8)
			       + ((u32)lut->sat[pelb >> 22] << 16);
			dst += 2;
		}
		dst += stride;
	}

	return 1;
}

/*  Delta-Huffman decompression routines  */
/*  translate_make_dehuff_lut()  */
#include "wnvhuff.h"


static void
translate_close(struct videum_device *dev)
{
	int	i;

	dev->translation1.type = XLAT_NULL;
	dev->translation1.in = NULL;
	dev->translation1.out = NULL;
	for (i = 0; i < 4; ++i)
	{
		dev->translation1.lut[i].type = LUT_NULL;
		if (dev->translation1.lut[i].table.base)
			vfree(dev->translation1.lut[i].table.base);
		dev->translation1.lut[i].table.base = NULL;
	}

	dev->translation2.type = XLAT_NULL;
	dev->translation2.in = NULL;
	dev->translation2.out = NULL;
	dev->translation2.lut[0].type = LUT_NULL;
	if (dev->translation2.lut[0].table.base)
		vfree(dev->translation2.lut[0].table.base);
	dev->translation2.lut[0].table.base = NULL;

	if (dev->xlat_temp)
		vfree(dev->xlat_temp);
	dev->xlat_temp = NULL;
}

static int
translate_setup(struct videum_device *dev)
{
	int	npix2;
	translate_close(dev);

	/*  Translation 1: Capture format to YUYV */
	dev->translation1.width = dev->uc_parm[UC_NORMAL].capture_width;
	dev->translation1.height = dev->uc_parm[UC_NORMAL].capture_height;
	dev->translation1.in_stride = 0;
	if (dev->capture_compress &&
	    dev->clientfmt.pixelformat == V4L2_PIX_FMT_WNVA)
	{
		dev->translation1.type = XLAT_NULL;
		dev->translation1.output_size = 0;
	}
	else if (dev->capture_compress)
	{
		dev->translation1.type = XLAT_HUFF_TO_YUYV;
		dev->translation1.out_stride = dev->translation1.width * 2;
		dev->translation1.output_size = dev->translation1.out_stride
			* dev->translation1.height;
		if (!translate_make_dehuff_lut(&dev->translation1))
			return 0;
		if (dev->xlat_temp != NULL)
			vfree(dev->xlat_temp);
		dev->xlat_temp = vmalloc(dev->translation1.output_size);
		if (dev->xlat_temp == NULL)
		{
			err_msg("Can't allocate xlat temp buffer %d bytes\n",
				dev->translation1.output_size);
			return 0;
		}
	}
	else
	{
		dev->translation1.type = XLAT_NULL;
		dev->translation1.output_size = 0;
	}

	/*  Translation 2: YUYV to client format, or NOP if WNVA */
	dev->translation2.width = dev->uc_parm[UC_NORMAL].capture_width;
	dev->translation2.height = dev->uc_parm[UC_NORMAL].capture_height;
	dev->translation2.in_stride = dev->translation2.width * 2;
	dev->translation2.output_size = dev->clientfmt.sizeimage;

	npix2 = dev->translation2.width * dev->translation2.height;
	switch (dev->clientfmt.pixelformat)
	{
	case V4L2_PIX_FMT_YUYV:
	case V4L2_PIX_FMT_WNVA:
		dev->translation2.type = XLAT_NULL;
		break;
	case V4L2_PIX_FMT_GREY:
		dev->translation2.type = XLAT_YUYV_TO_GREY;
		dev->translation2.out_stride = dev->translation2.width;
		break;
	case V4L2_PIX_FMT_YUV420:
		dev->translation2.type = XLAT_YUYV_TO_YUV420;
		dev->translation2.out_stride = dev->translation2.width;
		break;
	case V4L2_PIX_FMT_RGB555:
	case V4L2_PIX_FMT_RGB565:
		dev->translation2.type = (dev->clientfmt.pixelformat == 
			V4L2_PIX_FMT_RGB555) ? XLAT_YUYV_TO_RGB555 
			: XLAT_YUYV_TO_RGB565;
		dev->translation2.out_stride = dev->translation2.width * 2;
		if (!translate_make_rgb16_lut(&dev->translation2))
			return 0;
		break;
	case V4L2_PIX_FMT_BGR24:
		dev->translation2.type = XLAT_YUYV_TO_RGB24;
		dev->translation2.out_stride = dev->translation2.width * 3;
		if (!translate_make_rgb24_lut(&dev->translation2))
			return 0;
		break;
	case V4L2_PIX_FMT_BGR32:
		dev->translation2.type = XLAT_YUYV_TO_RGB32;
		dev->translation2.out_stride = dev->translation2.width * 4;
		if (!translate_make_rgb24_lut(&dev->translation2))
			return 0;
		break;
	}
	return 1;
}

static void
translate_inandout(struct videum_device *dev,
		   __u8 *input_buffer,
		   __u8 *output_buffer,
		   __u8 output_is_user_space)
{
	/*  Translation 1: Capture format to YUYV */
	dev->translation1.in = input_buffer;
	dev->translation1.output_is_user = 0;
	if (dev->translation1.type == XLAT_NULL)
	{
		dev->translation1.out = input_buffer;
	}
	else if (dev->capture_compress)
	{
		if (dev->capture_vdf == WAVI_VDF_YCQ)
			dev->translation1.type = XLAT_YCQ_TO_YUYV;
		else
		{
			dev->translation1.type = XLAT_HUFF_TO_YUYV;
			dev->translation1.ilut = (dev->capture_vdf >> 5) - 4;
		}
		dev->translation1.out = dev->xlat_temp;
	}

	/*  Translation 2: YUYV to client format, or NOP for WNVA */
	dev->translation2.in = dev->translation1.out;
	dev->translation2.out = output_buffer;
	dev->translation2.output_is_user = output_is_user_space;
}

static int /* length of output image or negative error */
translate_image(struct videum_device *dev,
		__u8	*input_buffer,
		__u8	*output_buffer,
		int	len,
		int	output_is_user)
{
	int	err;

	/* The buffer must be large enough for the whole image */
	if (len < dev->translation2.output_size)
	{
		debug_msg("Read buffer too small, %d < %d\n",
			  len, dev->translation2.output_size);
		return -EFAULT;
	}
	if (len > dev->translation2.output_size)
		len = dev->translation2.output_size;

	translate_inandout(dev, input_buffer, output_buffer, output_is_user);

	/*  Translation 1: Capture format to YUYV */
	if (dev->capture_compress)
	{
		if (dev->translation1.out == NULL)
		{
			err_msg("Internal error, xlat_temp == NULL\n");
			return -EFAULT;
		}
		switch (dev->translation1.type)
		{
		case XLAT_YCQ_TO_YUYV:
			translate_ycq_yuyv(&dev->translation1);
			break;
		case XLAT_HUFF_TO_YUYV:
			translate_dh_yuyv(&dev->translation1);
			break;
		}
	}


	/*  Translation 2: YUYV to client format, or NOP for WNVA */
	if (dev->translation2.in == dev->translation2.out)
	{
		return len;
	}
	if (dev->translation2.type == XLAT_NULL)
	{
		if (!output_is_user)
		{
			memcpy(output_buffer, dev->translation2.in, len);
		}
		else
		{
			err = copy_to_user(output_buffer, 
					   dev->translation2.in, len);
			len = (err) ? -EFAULT : len;
		}
		return len;
	}

	if (output_is_user && !access_ok(VERIFY_WRITE, output_buffer, len))
		return -EFAULT;

	switch (dev->translation2.type)
	{
	case XLAT_YUYV_TO_GREY:
		translate_yuyv_grey(&dev->translation2);
		break;
	case XLAT_YUYV_TO_YUV420:
		translate_yuyv_yuv420(&dev->translation2);
		break;
	case XLAT_YUYV_TO_RGB555:
	case XLAT_YUYV_TO_RGB565:
		translate_yuyv_rgb16(&dev->translation2);
		break;
	case XLAT_YUYV_TO_RGB24:
		translate_yuyv_rgb24(&dev->translation2);
		break;
	case XLAT_YUYV_TO_RGB32:
		translate_yuyv_rgb32(&dev->translation2);
		break;
	}
	dev->translation2.out = NULL;
	return len;
}

/*
 *
 *	V I D E O   C A P T U R E   F U N C T I O N S
 *
 */

/*
 *  Supported capture formats
 */
static struct v4l2_fmtdesc capfmt[] = 
{
#ifdef CONFIG_WNV_RGB
	{	0, {"RGB-16 (5-5-5)"},
		V4L2_PIX_FMT_RGB555,  0, 16, {0, 0},
	},
	{	1, {"RGB-16 (5-6-5)"},
		V4L2_PIX_FMT_RGB565,  0, 16, {0, 0},
	},
	{	2, {"RGB-24 (B-G-R)"},
		V4L2_PIX_FMT_BGR24,   0, 24, {0, 0},
	},
	{	3, {"RGB-32 (B-G-R-?)"},
		V4L2_PIX_FMT_BGR32,   0, 32, {0, 0},
	},
#endif
	{	4, {"Greyscale-8"},
		V4L2_PIX_FMT_GREY,    V4L2_FMT_CS_601YUV, 8, {0, 0},
	},
	{	5, {"YUV 4:2:2 (Y-U-Y-V)"},
		V4L2_PIX_FMT_YUYV,    V4L2_FMT_CS_601YUV, 16, {0, 0},
	},
	{	6, {"YUV 4:2:0 (planar)"},
		V4L2_PIX_FMT_YUV420,  V4L2_FMT_CS_601YUV, 12, {0, 0},
	},
	{	6, {"Winnov H/W Compressed"},
		V4L2_PIX_FMT_WNVA,  V4L2_FMT_CS_601YUV | 
		V4L2_FMT_FLAG_COMPRESSED, 8, {0, 0},
	},
};
#define NUM_CAPFMT (sizeof(capfmt)/sizeof(capfmt[0]))

static void
capture_limit_dimensions(struct videum_device *dev,
			 int	max_pixels,
			 int	*width,
			 int	*height)
{
	int	t;

	t = isqrt((max_pixels * dev->clientfmt.width) / dev->clientfmt.height);
	if (t < *width)
		*width = t;
	t = isqrt((max_pixels * dev->clientfmt.height) / dev->clientfmt.width);
	if (t < *height)
		*height = t;
	*width &= ~3;
	*height &= ~3;
}

/*  Stop the music!
 */
static void
capture_abort(struct videum_device *dev)
{
	dev->sampler_enabled = 0;
	wavi_video_sampler(dev, 0);
	wnv_bm_disarm(dev);
}

/*  The image format has changed, width, height, pixel format.
 *  Decide if the format is ok or take the closest valid format.
 */
static void
capture_new_format(struct videum_device *dev)
{
	int	ntsc;
	int	max_image_size;
	int	max_height;
	int	max_width;
	int	max_pixels;

	capture_abort(dev);

	/* BUG: won't work for auto */
	ntsc = (dev->videc.standard == V4L2_STD_NTSC);
	dev->ready_to_capture = 0;
	dev->uc_parm[UC_NORMAL].source_x = 
		(ntsc) ? dev->videc.ntsc_hskip : dev->videc.pal_hskip;
	dev->uc_parm[UC_NORMAL].source_y = 
		(ntsc) ? dev->videc.ntsc_vskip : dev->videc.pal_vskip;
	dev->uc_parm[UC_NORMAL].source_width = 
		(ntsc) ? dev->videc.ntsc_width : dev->videc.pal_width;
	dev->uc_parm[UC_NORMAL].source_height = 
		(ntsc) ? dev->videc.ntsc_height : dev->videc.pal_height;
	dev->uc_parm[UC_NORMAL].field = dev->videc.preferred_field + 1;

	dev->clientfmt.flags = V4L2_FMT_CS_601YUV;
	switch (dev->clientfmt.pixelformat)
	{
#ifndef CONFIG_WNV_RGB
	default:
#endif
	case V4L2_PIX_FMT_YUYV:
		dev->clientfmt.depth = 16;
		dev->clientfmt.pixelformat = V4L2_PIX_FMT_YUYV;
		break;
	case V4L2_PIX_FMT_GREY:
		dev->clientfmt.depth = 8;
		break;
	case V4L2_PIX_FMT_YUV420:
		dev->clientfmt.depth = 12;
		break;
	case V4L2_PIX_FMT_WNVA:
		dev->clientfmt.depth = 8;
		dev->clientfmt.flags |= V4L2_FMT_FLAG_COMPRESSED;
		break;
#ifdef CONFIG_WNV_RGB
	case V4L2_PIX_FMT_RGB555:
	case V4L2_PIX_FMT_RGB565:
		dev->clientfmt.depth = 16;
		dev->clientfmt.flags = 0;
		break;
	case V4L2_PIX_FMT_BGR24:
		dev->clientfmt.depth = 24;
		dev->clientfmt.flags = 0;
		break;
	case V4L2_PIX_FMT_BGR32:
		dev->clientfmt.depth = 32;
		dev->clientfmt.flags = 0;
		break;
	default:
		debug_msg("Unknown format \"%4.4s\" %08X\n", 
			  (char *)&dev->clientfmt.pixelformat,
			  dev->clientfmt.pixelformat);
		dev->clientfmt.depth = 16;
		dev->clientfmt.pixelformat = V4L2_PIX_FMT_RGB565;
		dev->clientfmt.flags = 0;
		//return;
		break;
#endif
	}
//debug_msg("Format \"%4.4s\"\n",(char *)&dev->clientfmt.pixelformat);

	if (dev->clientfmt.width < MIN_WIDTH)
		dev->clientfmt.width = MIN_WIDTH;
	if (dev->clientfmt.height < MIN_HEIGHT)
		dev->clientfmt.height = MIN_HEIGHT;

	/*  Compute max number of pixels we can capture, and limit the	*/
	/*  client width and height to that limit.			*/

	/*  Initial limit from hardware decoder  */
	max_width = dev->uc_parm[UC_NORMAL].source_width;
	max_height = dev->uc_parm[UC_NORMAL].source_height;
	if (dev->clientfmt.width > max_width)
		dev->clientfmt.width = max_width;
	if (dev->clientfmt.height > max_height)
		dev->clientfmt.height = max_height;
	dev->clientfmt.width &= ~3;
	dev->clientfmt.height &= ~3;

	/*  If stream buffers are mapped, the maximum number of pixels  */
	/*  we can capture may be limited by the buffer length          */
	if (dev->stream_buffers_mapped)
	{
		int	client_bpp = dev->clientfmt.depth;
		max_pixels = 8 * dev->stream_buf[0].vidbuf.length
			/ client_bpp;
		capture_limit_dimensions(dev, max_pixels, 
					 &dev->clientfmt.width,
					 &dev->clientfmt.height);
	}

	/*  Decide compressed or uncompressed. Use uncompressed if there  */
	/*  is enough SRAM memory. Use compressed for WNVA.  */
	max_image_size = dev->sram_video_size;
	dev->capture_compress = 0;
	dev->capture_bpp = 16;
	max_pixels = 8 * max_image_size / dev->capture_bpp;
	if (max_pixels <= dev->clientfmt.width * dev->clientfmt.height ||
	    dev->clientfmt.pixelformat == V4L2_PIX_FMT_WNVA)
	{/*	Compress!  */
		dev->capture_marker_spacing = 1024;
		dev->capture_compress = 1;
		dev->capture_bpp = 4;
		dev->capture_size = dev->sram_video_size
			& ~(dev->capture_marker_spacing - 1);
		max_pixels = 8 * dev->capture_size / dev->capture_bpp;
	}
	/*  Limit dimensions to how many pixels the hardware can fit	*/
	/*  in the SRAM.						*/
	capture_limit_dimensions(dev, max_pixels, 
				 &dev->clientfmt.width,
				 &dev->clientfmt.height);

	dev->clientfmt.sizeimage = (dev->clientfmt.width
		* dev->clientfmt.height
		* dev->clientfmt.depth)
		/ 8;

	/*  Skip a small offset into the buffer to place the image data
	 *  to allow room for a data header. Used only for WNVA compression
	 */
	dev->capture_offset = 0;
	if (dev->clientfmt.pixelformat == V4L2_PIX_FMT_WNVA)
		dev->capture_offset = sizeof(struct wnva_header);
	dev->clientfmt.sizeimage += dev->capture_offset;

	if (!dev->capture_compress)
	{
		dev->capture_size = dev->clientfmt.width
			* dev->clientfmt.height
			* dev->capture_bpp
			/ 8;
		dev->capture_min_size = dev->capture_size;
		dev->capture_vdf = WAVI_VDF_YCH;
	}
	if (dev->capture_compress)
	{
		dev->capture_min_size = dev->clientfmt.width
			* dev->clientfmt.height
			/ 4;  /* Two bits per pixel  */
		dev->capture_wcase_size = dev->clientfmt.width
			* dev->clientfmt.height
			/ 2;  /* Four bits per pixel  */
		dev->capture_vdf = WAVI_VDF_DH5;
		dev->capture_low_water = LOW_WATER * dev->capture_size / 100;
		dev->capture_high_water = HIGH_WATER * dev->capture_size / 100;
	}

	/*  Normal microcode */
	dev->uc_parm[UC_NORMAL].capture_width = dev->clientfmt.width;
	dev->uc_parm[UC_NORMAL].capture_height = dev->clientfmt.height;
	dev->uc_parm[UC_NORMAL].capture_min_size = dev->capture_min_size;
	dev->uc_parm[UC_NORMAL].capture_max_size = dev->capture_size;

	/*  Worstcase microcode */
	if (dev->capture_compress)
	{
		memcpy(&dev->uc_parm[UC_WORSTCASE], &dev->uc_parm[UC_NORMAL], 
		       sizeof(dev->uc_parm[0]));
		dev->uc_parm[UC_WORSTCASE].capture_height =
			dev->clientfmt.height >> 1;
		dev->uc_parm[UC_WORSTCASE].capture_min_size =
			dev->capture_wcase_size;
		dev->uc_parm[UC_WORSTCASE].capture_max_size =
			dev->capture_wcase_size;
	}
	rebuild_mmap_scatter_lists(dev);
}

static void
capture_armdma(struct videum_device *dev,
	       struct scatter_node *list)
{
	if (dev->dma_mode == DMAMODE_SLAVE)
		return;
	wavi_writereg(dev, WAVI_MXRPTR, dev->sram_video >> 4);
	wnv_bm_arm(dev, list, dev->clientfmt.pixelformat == V4L2_PIX_FMT_UYVY);
}

static void
capture_dma_line(struct videum_device	*dev,
		 struct ucode_parms	*parm)
{
	int	height	= parm->source_height;
	int	rate;
	int	line;

	parm->dma_line = 0;
	rate = dev->dma_rate / 15625;/* bytes per 64us */
	if (dev->dma_mode == DMAMODE_SLAVE || rate == 0 || height == 0)
		return;

	/* number of horizontal periods to read frame */
	line = parm->capture_min_size / rate;
	line = height - line + 5;
	if (line < 5)
		line = 5;
	if (line > height)
		line = height;
	//debug_msg("DMA line %d\n", line);
	parm->dma_line = line;
}

static void
capture_int_line(struct videum_device	*dev,
		 struct ucode_parms	*parm)
{
	int	height	= parm->source_height;
	int	rate;
	int	line	= height;

	parm->int1_line = height;
	if (height == 0)
		return;
	if (dev->dma_mode == DMAMODE_SLAVE)
	{/*  Non-DMA case */
		if (dev->slave_rate == 0)
			return;
		rate = dev->slave_rate / 15625;
		line = parm->capture_min_size / rate;
		line = height - line;
		if (line < 1)
			line = 1;
		if (line > height)
			line = height;
	}
	if (dev->dma_mode == DMAMODE_SIMPLE)
	{
		if (dev->dma_rate == 0)
			return;
		rate = dev->dma_rate / 15625;
		line = parm->capture_max_size / rate;
		line = (line * (100 + slowdma * 10)) / 100;/* fudge dma time */
		line = parm->dma_line + line;
		if (line <= parm->dma_line)
			line = parm->dma_line + 5;
		if (line > 1000)
			line = 1000;
	}
	//debug_msg("Interrupt line %d\n",line);
	parm->int1_line = line;
}

static int
compile_motion_microcode(struct videum_device *dev,
			 int	uc)
{
	capture_dma_line(dev, &dev->uc_parm[uc]);
	if (dev->uc_parm[uc].dma_line == 0)
	{
		//debug_msg("DMA line zero\n");
		dev->dma_mode = DMAMODE_SLAVE;
	}
	capture_int_line(dev, &dev->uc_parm[uc]);
	dev->uc_parm[uc].int2_line = 0;

	dev->uc_parm[uc].flags =  0;
	if (dev->type == VIDEUM_TYPE_PCI && dev->dma_mode != DMAMODE_SLAVE)
		dev->uc_parm[uc].flags |= 
			COMPFLAG_AVPRO_DMA | COMPFLAG_AVPRO_INT;
	else
		dev->uc_parm[uc].flags |= COMPFLAG_VST_WHEN_DONE;
	if (dev->capture_compress && uc == UC_NORMAL)
		dev->uc_parm[uc].flags |= COMPFLAG_HUFFMAN;

	dev->uc_loaded = UC_UNDEF;
	dev->uc_parm[uc].buffer = dev->uc_buffer[uc];
	dev->uc_length[uc] =
		compile_microcode(dev, &dev->uc_parm[uc]);
	if (dev->uc_length[uc] == 0)
	{/* ucode too long, no vertical filter reduces the length */
		dev->uc_parm[uc].flags |= COMPFLAG_NOVERTFILTER;
		dev->uc_length[uc] = compile_microcode(
			dev, &dev->uc_parm[uc]);
		if (dev->uc_length[uc] == 0)
		{
			err_msg("Internal error - can't compile ucode\n");
			return 0;/* failed */
		}
	}
	return 1;/* succeeded */
}

/*  Allocate buffers, compile microcode, and get everything ready to capture
 *  an image, but don't start capturing yet.
 */
static int
capture_begin(struct videum_device *dev)
{
	u32	min_capture_buffer_size;

	capture_abort(dev);
	if (dev->ready_to_capture)
		return dev->ready_to_capture;

	min_capture_buffer_size = dev->capture_size + dev->capture_offset;
	if (dev->capture_buffer_size < min_capture_buffer_size)
	{
		if (dev->capture_buffer != NULL)
			vfree(dev->capture_buffer);
		dev->capture_buffer_size = 
			(min_capture_buffer_size + PAGE_SIZE - 1) & 
			~(PAGE_SIZE - 1);
		dev->capture_buffer = (__u8 *)
			vmalloc(dev->capture_buffer_size);
		if (dev->capture_buffer == NULL)
		{
			dev->capture_buffer_size = 0;
			err_msg("Can't allocate capture buffer"
				" %d bytes\n", dev->capture_size);
			return dev->ready_to_capture;
		}
	}

	dev->dma_mode = DMAMODE_SLAVE;
	if (dev->irq && dev->dma_rate &&
	    wnv_bm_build_scatter_list(dev, dev->capture_buffer, 
				      &dev->capture_dma_list))
	{
		dev->dma_mode = DMAMODE_SIMPLE;
	}

	if (!compile_motion_microcode(dev, UC_NORMAL))
		return dev->ready_to_capture;
	if (dev->capture_compress &&
	    !compile_motion_microcode(dev, UC_WORSTCASE))
		return dev->ready_to_capture;

	if (!translate_setup(dev))
		return dev->ready_to_capture;

	//debug_msg("Ready to capture.  DMA mode = %d\n",dev->dma_mode);
	wnv_interrupt_enable(dev);
	return (dev->ready_to_capture = 1);
}

/*  Start an image capture
 */
static void
capture_grab_frame(struct videum_device *dev)
{
	struct scatter_node	*list;
	int			uc	= UC_NORMAL;

	if (dev->ready_to_capture && dev->sampler_enabled)
		return;

	capture_begin(dev);
	if (!dev->ready_to_capture)
		return;

	if (dev->capture_vdf == WAVI_VDF_YCQ)
		uc = UC_WORSTCASE;
	if (dev->uc_loaded != uc)
	{
		if (dev->type == VIDEUM_TYPE_PCI)
			wavi_do_rvs(dev);
		wavi_block_write(dev, dev->sram_ucode,
				 dev->uc_buffer[uc],
				 dev->uc_length[uc], 1);
		wavi_filter(dev, dev->uc_parm[uc].filter);
		dev->uc_loaded = uc;
	}
	wavi_vdf(dev, dev->capture_vdf);

	wavi_writereg(dev, WAVI_SPLPTR, dev->sram_ucode >> 4);
	wavi_writereg(dev, WAVI_VID2PTR, dev->sram_linebuf >> 4);
	wavi_writereg(dev, WAVI_VID1PTR, dev->sram_video >> 4);

	if (dev->capture_vdf >= WAVI_VDF_DH4)
	{
		int	n;
		int	start;
		int	msm1	= dev->capture_marker_spacing - 1;
		start = (dev->capture_min_size + msm1) & ~msm1;
		n = ((dev->capture_size + msm1) & ~msm1)
			- start;
		wavi_write_markers(dev, dev->sram_video + start,
				   n >> 10, dev->capture_marker_spacing);
	}

	list = dev->capture_dma_list;/* DMA list for capture_buffer */
	if (dev->streaming)
	{
		struct stream_buffer	*buf;

		/*  Capture to capture_buffer  */
		dev->stream_capture_buffer = dev->capture_buffer;

		/*  Capture straight into streaming buffer? */
		if (dev->translation1.type == XLAT_NULL &&
		    dev->translation2.type == XLAT_NULL)
		{
//			dev->stream_capture_buffer = NULL;/*test:*/
			buf = v4l2_q_peek_head(&dev->stream_q_capture);
			if (buf != NULL)
			{
				dev->stream_capture_buffer = buf->vaddress;
				list = buf->dma_list;
			}
//			if (buf == NULL)/*test:*/
//				return;
		}
	}

	if (dev->dma_mode != DMAMODE_SLAVE)
		capture_armdma(dev, list);
	wavi_video_sampler(dev, 1);
	dev->sampler_enabled = 1;
	dev->capture_completed = 0;
}

/*
 *	STREAMING CAPTURE
 */

static int/* 1 = success; 0 = failed */
capture_queuebuffer(struct videum_device *dev,
		    struct v4l2_buffer	 *vidbuf)
{
	int			i	= vidbuf->index;
	struct stream_buffer	*buf	= NULL;

	if (!dev->stream_buffers_mapped || 
	    vidbuf->type != V4L2_BUF_TYPE_CAPTURE)
	{
		debug_msg("QBUF no buffers mapped or wrong type\n");
		return 0;
	}
	if (i < 0 || i >= MAX_CAPTURE_BUFFERS || !dev->stream_buf[i].requested)
	{
		debug_msg("QBUF buffer index %d is out of range\n", i);
		return 0;
	}

	buf = &dev->stream_buf[i];

	if (!(buf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED))
	{
		debug_msg("QBUF buffer %d is not mapped\n", i);
		return 0;
	}
	if ((buf->vidbuf.flags & V4L2_BUF_FLAG_QUEUED))
	{
		debug_msg("QBUF buffer %d is already queued\n", i);
		return 0;
	}

	buf->vidbuf.flags &= ~V4L2_BUF_FLAG_DONE;
	v4l2_q_add_tail(&dev->stream_q_capture, &buf->qnode);
	buf->vidbuf.flags |= V4L2_BUF_FLAG_QUEUED;
	if (dev->streaming)
		capture_grab_frame(dev);/* does nothing if already capturing */
	return 1;
}

static int/* 1 = got a buffer; 0 = no buffers */
capture_dequeuebuffer(struct videum_device *dev,
		      struct v4l2_buffer *buf)
{
	struct stream_buffer *newbuf;

	if (buf->type != V4L2_BUF_TYPE_CAPTURE)
	{
		debug_msg("DQBUF wrong buffer type\n");
		return 0;
	}
	newbuf = v4l2_q_del_head(&dev->stream_q_done);
	if (newbuf == NULL)
	{
		debug_msg("DQBUF nothing on done queue\n");
		return 0;
	}
	newbuf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;

	*buf = newbuf->vidbuf;
	return 1;
}

static int
capture_streamon(struct videum_device	*dev,
		 __u32			type)
{
	struct stream_buffer *buf;

	if (type != V4L2_BUF_TYPE_CAPTURE)
		return 0;
	if (dev->streaming)
	{
		capture_grab_frame(dev);
		return 1;
	}

	capture_abort(dev);/* cancel any capture that might be in progress */

	/*  -2 is a magic number that triggers start-of-stream logic in */
	/*    capture_interrupt()  */
	dev->stream_last_frame = -2;

	dev->perf.frames = 0;
	dev->perf.framesdropped = 0;
	dev->perf.bytesout = 0;

	/*  Can't capture frames faster than the video input  */
	if (dev->capture.timeperframe < dev->videc.frame_period)
		dev->capture.timeperframe = dev->videc.frame_period;

	/*  Move any leftover DONE buffers to the free pool */
	while ((buf = v4l2_q_del_head(&dev->stream_q_done)))
		buf->vidbuf.flags &= ~V4L2_BUF_FLAG_QUEUED;

	/*  Kick off the machine */
	dev->streaming = 1;
	capture_grab_frame(dev);
	return 1;
}

static void
capture_streamoff(struct videum_device	*dev,
		 __u32			type)
{
	if (!dev->streaming || type != V4L2_BUF_TYPE_CAPTURE)
		return;
	capture_abort(dev);
	dev->streaming = 0;

	/* Note: should really delay this till next capture */
	dev->perf.frames = 0;
	dev->perf.framesdropped = 0;
	dev->perf.bytesout = 0;
}

/*  Add header for WNVA  */
static void
capture_wnva_header(struct videum_device *dev,
		    u8 *buffer)
{
	struct wnva_header *hdr;

	if (dev->clientfmt.pixelformat != V4L2_PIX_FMT_WNVA)
		return;

	hdr = (struct wnva_header *)buffer;
	hdr->width = dev->uc_parm[UC_NORMAL].capture_width;
	hdr->height = dev->uc_parm[UC_NORMAL].capture_height;
	hdr->flags = dev->capture_vdf >> 5;
	if (dev->capture_vdf == WAVI_VDF_YCQ)
	{
		hdr->width = dev->uc_parm[UC_WORSTCASE].capture_width;
		hdr->height = dev->uc_parm[UC_WORSTCASE].capture_height;
	}
}

/*	Read out and convert the next frame
 */
static int /* returns length of data or negative for error */
capture_imagereadout(struct videum_device	*dev,
		     __u8			*capture_buffer,
		     __u8			*output_buffer,
		     int			output_size,
		     int			output_is_user)
{
	int	len		= dev->capture_size;
	int	native_len;

	capture_buffer += dev->capture_offset;
	output_buffer += dev->capture_offset;

	if (dev->capture_vdf == WAVI_VDF_YCQ)
		len = dev->capture_wcase_size;

	if (dev->dma_mode == DMAMODE_SLAVE && dev->capture_vdf < WAVI_VDF_DH4)
	{
		int	i, n, ram, block = 1024;
		__u8	*p;
		/*  Remember that ints are disabled in block_read() */
		n = (len + block - 1) / block;
		p = capture_buffer;
		ram = dev->sram_video;
		for (i = 0; i < n; ++i, ram += block, p += block)
			wavi_block_read(dev, ram, p, block);
		if (!wavi_vst(dev))
		{
			debug_msg("Slave readout underrun!\n");
			len = -1;/* don't copy to user */
		}
	}
	if (dev->dma_mode == DMAMODE_SLAVE && dev->capture_vdf >= WAVI_VDF_DH4)
	{
		int	ms	= dev->capture_marker_spacing;
		int	i, ram	= dev->sram_video;
		u32	*p	= (u32 *)capture_buffer;

		/*  Find compressed length  */
		len = 0;
		for (i = 0; i < dev->capture_size; i += ms)
		{
			wavi_block_read(dev, ram, (__u8 *)p, ms);
			ram += ms;
			(char *)p += ms;
			if (p[-4] == 0xffffffff &&
			    p[-3] == 0xffffffff &&
			    p[-2] == 0xffffffff &&
			    p[-1] == 0xffffffff)
			{
				len = i + ms;
				break;
			}
		}
		if (!wavi_vst(dev))
		{
			debug_msg("Slave readout underrun!\n");
			len = -1;/* don't copy to user */
		}
	}
	if (dev->dma_mode == DMAMODE_SIMPLE && dev->capture_vdf < WAVI_VDF_DH4)
	{
		if (!wnv_bm_is_done(dev))
		{
			//debug_msg("DMA not finished. state: %d\n",
			//	  dev->dma_state);
			/* discard frame and try again */
			len = -1;/* don't copy to user */
		}
		else
		{
			//debug_msg("DMA is done.\n");
		}
		wnv_bm_resync(dev);
	}
	if (dev->dma_mode == DMAMODE_SIMPLE &&
	    dev->capture_vdf >= WAVI_VDF_DH4)
	{
		int	i, ms	= dev->capture_marker_spacing;
		u32	*p	= (u32 *)(capture_buffer + ms);

		if (!wnv_bm_is_done(dev))
		{
			//debug_msg("DMA not finished 2\n");
			/* discard frame and try again */
			len = -1;/* don't copy to user */
		}
		/*  Find compressed length  */
		len = 0;
		for (i = ms; i <= dev->capture_size; i += ms)
		{
			if (p[-4] == 0xffffffff &&
			    p[-3] == 0xffffffff &&
			    p[-2] == 0xffffffff &&
			    p[-1] == 0xffffffff)
			{
				len = i;
				break;
			}
			(char *)p += ms;
		}
		wnv_bm_resync(dev);
	}
	dev->sampler_enabled = 0;
	wavi_video_sampler(dev, 0);

	if (len <= 0)
	{
		dev->capture_vdf =
			compression_adjust(dev, dev->capture_vdf, len);
		return 0;
	}
	native_len = len;

	len = translate_image(dev, capture_buffer, output_buffer,
			      output_size, output_is_user);
	output_buffer -= dev->capture_offset;

	if (len > 0)
		capture_wnva_header(dev, output_buffer);

	dev->capture_vdf = compression_adjust(dev, dev->capture_vdf,
					      native_len);
	if (len < 0)
	{
		return len;
	}
	++dev->perf.frames;
	dev->perf.bytesout += len;
	return len;
}

/*  The hardware has issued the interrupt signal, do any post-capture
 *  processing that may be necessary.
 *
 *  This function is called from an interrupt bottom-half handler.
 */
static void
capture_interrupt(void *v)
{
	struct videum_device	*dev = (struct videum_device *)v;
	int			itr;
	struct stream_buffer	*buf;
	int			len;
	stamp_t			timestamp_rough;
	unsigned long		raw_frame_num;
	unsigned long		next_raw_frame_to_keep;
	unsigned long		stream_frame_num;
	stamp_t			temp64;

	itr = wavi_readreg(dev, WAVI_ITR) & (WAVI_VI1BIT | WAVI_VI2BIT);
	if (itr == 0)
	{
		return;
	}
	if (dev->capture_buffer == NULL ||
	    !dev->sampler_enabled || !dev->ints_enabled)
	{
		err_msg("Interrupt handler aborted.\n");
		return;
	}
	dev->capture_completed = 1;

	if (!dev->streaming)
	{
		dev->time_acquired = current_time_ms();
		/* DMA might not have finished, but we'll check in read() */
		wake_up_interruptible(&dev->new_video_frame);
		return;
	}

	/*  Only get here in streaming mode  */

	if (dev->stream_last_frame == -2)
	{/*	Very first frame of stream  */
		v4l2_masterclock_gettime(&dev->stream_begin);
		dev->stream_last_frame = -1;
	}

	if (dev->stream_capture_buffer == NULL)
	{
		return;
	}

	buf = v4l2_q_peek_head(&dev->stream_q_capture);
	if (buf == NULL)
	{/*	No available buffers. Skip this frame */
		dev->sampler_enabled = 0;
		wavi_video_sampler(dev, 0);
		capture_grab_frame(dev);
		return;
	}

	if (dev->videc.frame_period == 0)
	{/*	Sanity check to avoid divide-by-zero  */
		err_msg("videc.frame_period == 0 -- divide by 0\n");
		dev->videc.frame_period = 333667;
	}

	v4l2_masterclock_gettime(&timestamp_rough);
	/*  We do computations on a relative timestamp (since beginning  */
	/*  of stream), then add 'dev->stream_begin' back in later       */
	timestamp_rough -= dev->stream_begin;

	raw_frame_num = v4l2_timestamp_divide(
		timestamp_rough, dev->videc.frame_period);
	temp64 = (stamp_t)dev->capture.timeperframe
		* (dev->stream_last_frame + 1)
		+ (dev->videc.frame_period >> 1);
	next_raw_frame_to_keep = 
		v4l2_math_div6432((u64)temp64, dev->videc.frame_period, NULL);
	if (raw_frame_num < next_raw_frame_to_keep)
	{/*	Not time yet, don't keep this frame */
		dev->sampler_enabled = 0;
		wavi_video_sampler(dev, 0);
		capture_grab_frame(dev);
		return;
	}

	len = capture_imagereadout(dev, 
		dev->stream_capture_buffer,
		buf->vaddress, 
		buf->vidbuf.length, 0);

	if (len <= 0)
	{/*	Frame no good, DMA did not finish, etc. */
		capture_grab_frame(dev);
		return;
	}
	
	buf->vidbuf.bytesused = len;
	buf->vidbuf.flags |= V4L2_BUF_FLAG_DONE | V4L2_BUF_FLAG_KEYFRAME;

	stream_frame_num = v4l2_timestamp_correct(&timestamp_rough,
						  dev->capture.timeperframe);
	buf->vidbuf.timestamp = timestamp_rough
		              + dev->stream_begin; /* make stamp absolute */
	buf->vidbuf.sequence = stream_frame_num;

	if (stream_frame_num > dev->stream_last_frame + 1)
	{/*	We have missed one or more frames  */
		dev->perf.framesdropped += stream_frame_num
			- dev->stream_last_frame - 1;
		//debug_msg("Dropped %d frames, %d total. Frame %d\n",
		//	  (int)(stream_frame_num-dev->stream_last_frame-1),
		//	  (int)dev->perf.framesdropped,(int)stream_frame_num);
	}
	dev->stream_last_frame = stream_frame_num;


	buf = v4l2_q_del_head(&dev->stream_q_capture);
	v4l2_q_add_tail(&dev->stream_q_done, &buf->qnode);

	capture_grab_frame(dev);
	wake_up_interruptible(&dev->new_video_frame);
}


/*  Read captured data into a user buffer.
 *  Return: negative = error
 *	    0        = keep waiting
 *	    positive = count of bytes read successfully
 */
static long
capture_read(struct videum_device *dev,
	     __u8	*user_buffer,
	     int	user_buffer_size)
{
	int		len = user_buffer_size;
	unsigned long	now;

	if (!dev->ints_enabled)
		return -EIO;

	if (!dev->capture_completed)
	{/* No interrupt has occurred yet, or DMA didn't finish.  */
		//debug_msg("No data ready.\n");
		if (!dev->sampler_enabled)
			capture_grab_frame(dev);
		return 0;/* caller should keep waiting */
	}
	
	now = current_time_ms();
	if (now - dev->time_acquired > MAX_FRAME_AGE)
	{/* Frame in buffer is stale, get a new one */
		//debug_msg("Stale frame, re-acquiring.\n");
		dev->sampler_enabled = 0;
		wavi_video_sampler(dev, 0);
		capture_grab_frame(dev);
		return 0;/* caller should keep waiting */
	}

	len = capture_imagereadout(dev, dev->capture_buffer,
				   user_buffer, user_buffer_size, 1);
	capture_grab_frame(dev);
	return len;
}

/*  Stop capturing and free all resources used for capture.
 */
static void
capture_close(struct videum_device *dev)
{
	if (dev->streaming)
		capture_streamoff(dev, V4L2_BUF_TYPE_CAPTURE);
	capture_abort(dev);
	dev->ready_to_capture = 0;
	translate_close(dev);
	if (dev->capture_dma_list)
		free_page((unsigned long)dev->capture_dma_list);
	dev->capture_dma_list = 0;
	if (dev->capture_buffer != NULL)
		vfree(dev->capture_buffer);
	dev->capture_buffer = NULL;
	dev->capture_buffer_size = 0;
}

/*
 *
 *	M E M O R Y   M A P P I N G
 *
 */

static struct stream_buffer *
mmap_stream_buffer_from_offset(struct videum_device *dev,
			       unsigned long pgoff)
{
	int		i;
	unsigned long	offset	= pgoff * PAGE_SIZE;

	for (i = 0; i < MAX_CAPTURE_BUFFERS; ++i)
		if (offset == dev->stream_buf[i].vidbuf.offset)
			return &dev->stream_buf[i];
	return NULL;
}

static int
mmap_request_buffers(struct videum_device *dev,
		     struct v4l2_requestbuffers *req)
{
	int	i;
	u32	buflen;
	u32	type;

	if (dev->stream_buffers_mapped)
		return 0;/* can't make requests if buffers are mapped */
	if (req->count < 1)
		req->count = 1;
	if (req->count > MAX_CAPTURE_BUFFERS)
		req->count = MAX_CAPTURE_BUFFERS;
	type = V4L2_BUF_TYPE_CAPTURE;
	dev->stream_contig_map = 0;
	if (req->type & V4L2_BUF_REQ_CONTIG)
	{
		dev->stream_contig_map = 1;
		req->type = type | V4L2_BUF_REQ_CONTIG;
	}

	buflen = (dev->clientfmt.sizeimage + PAGE_SIZE - 1)
		& ~(PAGE_SIZE - 1);

//debug_msg("Granting %d buffers\n",req->count);

	/*  Now initialize the buffer structures. Don't allocate the */
	/*  buffers until they're mapped. */
	for (i = 0; i < req->count; ++i)
	{
		dev->stream_buf[i].requested = 1;
		dev->stream_buf[i].vidbuf.index = i;
		dev->stream_buf[i].vidbuf.type = type;
		/* offsets must be multiples of PAGE_SIZE, and never zero */
		dev->stream_buf[i].vidbuf.offset = PAGE_SIZE * (i + 1);
		dev->stream_buf[i].vidbuf.length = buflen;
		dev->stream_buf[i].vidbuf.bytesused = 0;
		dev->stream_buf[i].vidbuf.flags = 0;
		dev->stream_buf[i].vidbuf.timestamp = 0;
		dev->stream_buf[i].vidbuf.sequence = 0;
		memset(&dev->stream_buf[i].vidbuf.timecode, 0, 
		       sizeof(struct v4l2_timecode));
	}
	for (i = req->count; i < MAX_CAPTURE_BUFFERS; ++i)
		dev->stream_buf[i].requested = 0;
	dev->stream_buffers_requested = req->count;

	return 1;
}

static void
mmap_unrequest_buffers(struct videum_device *dev)
{
	int	i;

	if (dev->stream_buffers_requested == 0 ||
	    dev->stream_buffers_mapped)
		return;
	for (i = 0; i < MAX_CAPTURE_BUFFERS; ++i)
		dev->stream_buf[i].requested = 0;
	dev->stream_buffers_requested = 0;
}

static void
mmap_vma_open(struct vm_area_struct *vma)
{
	struct videum_device *dev =
		videum_device_from_file(vma->vm_file);
	if (dev == NULL)
		return;
	//debug_msg("vma_open called\n");
	MOD_INC_USE_COUNT;
}

static void
mmap_vma_close(struct vm_area_struct *vma)
{
	struct videum_device *dev =
		videum_device_from_file(vma->vm_file);
	struct stream_buffer *buf =
		mmap_stream_buffer_from_offset(dev, vma->vm_pgoff);
	int	i, n = 1;

	if (dev->stream_contig_map)
	{/*	Unmap all the buffers in one stroke  */
		n = dev->stream_buffers_mapped;
		buf = &dev->stream_buf[0];
	}
	for (i = 0; i < n; ++i)
	{
		if (dev->streaming)
		{
			info_msg("Warning- munmap() called while streaming\n");
			capture_streamoff(dev, buf->vidbuf.type);
		}
		v4l2_q_yank_node(&dev->stream_q_capture, &buf->qnode);
		v4l2_q_yank_node(&dev->stream_q_done, &buf->qnode);

		if (buf->vaddress != NULL && i == 0)
			vfree(buf->vaddress);
		buf->vaddress = NULL;
		if (buf->dma_list)
			free_page((unsigned long)buf->dma_list);
		buf->dma_list = NULL;
		buf->vidbuf.flags = 0;
		//debug_msg("Buffer %d deallocated\n",(int)vma->vm_pgoff);
		++buf;
		if (dev->stream_buffers_mapped > 0)
			--dev->stream_buffers_mapped;
	}
	MOD_DEC_USE_COUNT;
}

static struct page *
mmap_vma_nopage(struct vm_area_struct *vma,
		unsigned long address, int write)
{
	struct videum_device	*dev;
	struct stream_buffer	*buf;
	unsigned long		offset_into_buffer;
	struct page		*page;
	int			n = 1;

	dev = videum_device_from_file(vma->vm_file);
	if (dev == NULL)
		return 0;
	if (dev->stream_contig_map)
	{
		buf = &dev->stream_buf[0];
		n = dev->stream_buffers_requested;
	}
	else
		buf = mmap_stream_buffer_from_offset(dev, vma->vm_pgoff);
	if (buf == NULL)
		return 0;
	offset_into_buffer = address - vma->vm_start;
	if (offset_into_buffer >= buf->vidbuf.length * n)
		return 0;
	page = v4l2_vmalloc_to_page(buf->vaddress + offset_into_buffer);
	if (page == NULL)
		return 0;

	atomic_inc(&page->count);
	return page;
}

static struct vm_operations_struct videum_vma_operations =
{
	mmap_vma_open, mmap_vma_close, mmap_vma_nopage,
};


/*
 *
 *	V I D E O   F O R   L I N U X   I N T E R F A C I N G
 *
 */

static int
videum_open(struct v4l2_device *v, int flags, void **idptr)
{
	struct videum_device *dev = (struct videum_device *)v;
	int	i, n;
	int	cap;

	for (i = 0, n = -1, cap = 0; i < MAX_OPENS; ++i)
	{
		if (!dev->open_data[i].isopen)
			n = i;/* available open_data structure */
		else if (!dev->open_data[i].noncapturing)
			cap = 1;/* another open is already capturing */
	}
	if (n == -1)/* No available open_data structures */
		return -EBUSY;

	if (flags & O_NONCAP)/*  Non-capturing open */
		dev->open_data[n].noncapturing = 1;
	else if (cap)
		return -EBUSY;
	else
	{
		dev->open_data[n].noncapturing = 0;
		++dev->capturing_opens;
		dev->perf.frames = 0;
		dev->perf.framesdropped = 0;
		dev->perf.bytesout = 0;
	}

	++dev->open_count;
	dev->open_data[n].isopen = 1;
	dev->open_data[n].dev = dev;
	*idptr = &dev->open_data[n];

	if (dev->open_count == 1)
	{
		wnv_register_video_interrupt(dev, capture_interrupt);
		wavi_benchmark(dev);
		dev->ready_to_capture = 0;/* benchmark changes parameters! */
		dev->capture_completed = 0;
		dev->sampler_enabled = 0;
		v4l2_q_init(&dev->stream_q_capture);
		v4l2_q_init(&dev->stream_q_done);
	}
	debug_msg("Open '%s' %s\n", dev->v.name, 
		  (flags&O_NONCAP)?"non-capturing":"");
	MOD_INC_USE_COUNT;
	return 0;
}

static void
videum_close(void *id)
{
	struct device_open *o = (struct device_open *)id;
	struct videum_device *dev = o->dev;

	if (!o->noncapturing)
	{
		--dev->capturing_opens;
	}
	debug_msg("Close '%s'\n", dev->v.name);
	o->isopen = 0;
        --dev->open_count;
	if (dev->open_count == 0)
	{
		wnv_interrupt_disable(dev);
		capture_close(dev);
		wnv_unregister_video_interrupt(dev);
	}
	MOD_DEC_USE_COUNT;
}

static long
videum_write(void		*id, 
	     const char		*buf, 
	     unsigned long	count, 
	     int		noblock)
{
	return -EINVAL;
}

static int
videum_ioctl(void		*id,
	     unsigned int	cmd,
	     void		*arg)
{
	struct device_open *o = (struct device_open *)id;
	struct videum_device *dev = o->dev;

//debug_msg("ioctl %d\n", _IOC_NR(cmd));
	switch(cmd)
	{
	case VIDIOC_QUERYCAP:
	{
		struct v4l2_capability *b = arg;
		strcpy(b->name, dev->v.name);
		b->type = V4L2_TYPE_CAPTURE;
		b->flags =  V4L2_FLAG_READ |
			   V4L2_FLAG_STREAMING |
			   V4L2_FLAG_SELECT;
		b->inputs = dev->videc.num_inputs;
		b->outputs = 0;
		b->audios = 0;
		b->maxwidth = MAX_WIDTH;
		b->maxheight = MAX_HEIGHT;
		b->minwidth = MIN_WIDTH;
		b->minheight = MIN_HEIGHT;
		b->maxframerate = 30;
		return 0;
	}

	case VIDIOC_ENUM_PIXFMT:
	{
		struct v4l2_fmtdesc *f = arg;
		if (f->index < 0 || f->index >= NUM_CAPFMT)
			return -EINVAL;
		*f = capfmt[f->index];
		return 0;
	}

	case VIDIOC_G_FMT:
	{
		struct v4l2_format *fmt = arg;
		if (fmt->type != V4L2_BUF_TYPE_CAPTURE)
			return -EINVAL;
		fmt->fmt.pix = dev->clientfmt;
		return 0;
	}

	case VIDIOC_S_FMT:
	{
		struct v4l2_format *fmt = arg;
		if (o->noncapturing)
			return -EPERM;
		if (fmt->type != V4L2_BUF_TYPE_CAPTURE)
			return -EINVAL;
		dev->clientfmt = fmt->fmt.pix;
		capture_new_format(dev);
		if (dev->streaming)
			capture_grab_frame(dev);
		fmt->fmt.pix = dev->clientfmt;
		return 0;
	}

	case VIDIOC_G_COMP:	return -EINVAL;
	case VIDIOC_S_COMP:	return -EINVAL;

	case VIDIOC_REQBUFS:
	{
		struct v4l2_requestbuffers *req = arg;
		if (o->noncapturing || dev->stream_buffers_mapped)
			return -EPERM;
		mmap_unrequest_buffers(dev);
		capture_begin(dev);
		if (!mmap_request_buffers(dev, req))
			return -EINVAL;
		return 0;
	}

	case VIDIOC_QUERYBUF:
	{
		struct v4l2_buffer *buf = arg;
		int	i;
		if (o->noncapturing)
			return -EPERM;
		i = buf->index;
		if (i < 0 || i >= MAX_CAPTURE_BUFFERS ||
		    !dev->stream_buf[i].requested ||
		    (buf->type & V4L2_BUF_TYPE_field) != 
		     (dev->stream_buf[i].vidbuf.type & V4L2_BUF_TYPE_field))
			return -EINVAL;
		*buf = dev->stream_buf[i].vidbuf;
		return 0;
	}

	case VIDIOC_QBUF:
	{
		struct v4l2_buffer *buf = arg;
		if (o->noncapturing)
			return -EPERM;
		if (!dev->stream_buffers_mapped)
			return -EINVAL;
		if (!capture_queuebuffer(dev, buf))
			return -EINVAL;
		return 0;
	}

	case VIDIOC_DQBUF:
	{
		struct v4l2_buffer *buf = arg;
		if (o->noncapturing)
			return -EPERM;
		if (!capture_dequeuebuffer(dev, buf))
			return -EINVAL;
		return 0;
	}

	case VIDIOC_STREAMON:
	{
		__u32	*type = arg;
		if (o->noncapturing)
			return -EPERM;
		if (!capture_streamon(dev, *type))
			return -EINVAL;
		return 0;
	}

	case VIDIOC_STREAMOFF:
	{
		__u32	*type = arg;
		if (o->noncapturing)
			return -EPERM;
		capture_streamoff(dev, *type);
		return 0;
	}

	case VIDIOC_ENUM_FBUFFMT:	return -EINVAL;
	case VIDIOC_G_FBUF:		return -EINVAL;
	case VIDIOC_S_FBUF:		return -EINVAL;
	case VIDIOC_G_WIN:		return -EINVAL;
	case VIDIOC_S_WIN:		return -EINVAL;
	case VIDIOC_PREVIEW:		return -EINVAL;

	case VIDIOC_G_PERF:
	{
		memcpy(arg, &dev->perf, sizeof(dev->perf));
		return 0;
	}

	case VIDIOC_G_INPUT:
	{
		memcpy(arg, &dev->input, sizeof(dev->input));
		return 0;
	}

	case VIDIOC_S_INPUT:
	{
		int	input = *((int *)arg);
		if (input < 0 || input >= dev->videc.num_inputs)
		{
			debug_msg("Input out of range %d\n", input);
			return -EINVAL;
		}
		if (input != dev->input)
		{
			dev->input = input;
			set_video_input(dev, input);

			/* Changing the input disrupts the capture, so  */
			/* kick it off again  */
			if (dev->sampler_enabled)
			{
				capture_abort(dev);
				capture_grab_frame(dev);
			}
		}
		return 0;
	}

	case VIDIOC_G_PARM:
	{
		struct v4l2_streamparm *sp = arg;
		if (sp->type != V4L2_BUF_TYPE_CAPTURE)
			return -EINVAL;
		sp->parm.capture = dev->capture;
		return 0;
	}

	case VIDIOC_S_PARM:
	{
		struct v4l2_streamparm *sp = arg;
		struct v4l2_captureparm *vp = &sp->parm.capture;

		if (sp->type != V4L2_BUF_TYPE_CAPTURE)
			return -EINVAL;
		if (vp->capturemode & ~dev->capture.capability)
			return -EINVAL;
		if (vp->timeperframe < 10000)
			return -EINVAL;
		if (vp->capturemode != dev->capture.capturemode &&
		    !o->noncapturing && dev->streaming)
			return -EINVAL;

		if (o->noncapturing)
			return 0;
		if (vp->capturemode != dev->capture.capturemode)
		{
			dev->capture.capturemode = vp->capturemode;
			capture_new_format(dev);
		}
		if ((vp->capturemode & V4L2_CAP_TIMEPERFRAME) &&
		    vp->timeperframe >= dev->videc.frame_period)
			dev->capture.timeperframe = vp->timeperframe;
		else
			dev->capture.timeperframe = dev->videc.frame_period;
		return 0;
	}

	case VIDIOC_G_STD:
	{
		struct v4l2_standard *std = arg;
		v4l2_video_std_construct(std, dev->videc.standard, 0);
		return 0;
	}

	case VIDIOC_S_STD:
	{
		struct v4l2_standard	*std = arg;
		int			id;
		if ((o->noncapturing && dev->capturing_opens))
			return -EPERM;
		id = v4l2_video_std_confirm(std);
		if (!((1 << id) & dev->videc.standards))
		{
			debug_msg("Bad standard: %u\n", (unsigned)id);
			return -EINVAL;
		}
		dev->videc.set_standard(dev, id);
		capture_new_format(dev);
		if (dev->streaming)
			capture_grab_frame(dev);
		return 0;
	}

	case VIDIOC_ENUMSTD:
	{
		struct v4l2_enumstd *estd = arg;
		__u32	b, i;
		if (estd->index < 0 || estd->index > 30)
			return -EINVAL;
		for (b = 1, i = 0; b < 32; ++b)
		{
			if (((1 << b) & dev->videc.standards) == 0)
				continue;
			if (i == estd->index)
			{
				v4l2_video_std_construct(&estd->std, b, 0);
				estd->inputs = (__u32)-1; /* all inputs */
				estd->outputs = 0;
				return 0;
			}
			++i;
		}
		return -EINVAL;
	}

	case VIDIOC_ENUMINPUT:
	{
		struct v4l2_input *vi = arg;
		if (vi->index < 0 || vi->index >= dev->videc.num_inputs)
			return -EINVAL;
		*vi = dev->source[vi->index].input;
		return 0;
	}

	case VIDIOC_QUERYCTRL:
	{
		struct v4l2_queryctrl	*qc = arg;
		int			i;
		i = find_vctrl(qc->id);
		if (i == -EINVAL)
		{
			qc->flags = V4L2_CTRL_FLAG_DISABLED;
			return 0;
		}
		if (i < 0)
			return -EINVAL;
		videum_control[i].category = qc->category;
		memcpy(videum_control[i].group, qc->group, 
		       sizeof(qc->group));
		*qc = videum_control[i];
		return 0;
	}

	case VIDIOC_G_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int			i;
		i = find_vctrl(vc->id);
		if (i < 0)
			return -EINVAL;
		vc->value = dev->source[dev->input].control[i];
		return 0;
	}

	case VIDIOC_S_CTRL:
	{
		struct v4l2_control	*vc = arg;
		int			i;
		i = find_vctrl(vc->id);
		if (i < 0)
			return -EINVAL;
		dev->source[dev->input].control[i] = vc->value;
		wavi_tone_controls(dev);
		return 0;
	}

	case VIDIOC_G_TUNER:	return -EINVAL;
	case VIDIOC_S_TUNER:	return -EINVAL;
	case VIDIOC_G_FREQ:	return -EINVAL;
	case VIDIOC_S_FREQ:	return -EINVAL;

	case VIDIOC_G_AUDIO:	return -EINVAL;
	case VIDIOC_S_AUDIO:	return -EINVAL;

	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

static int
videum_mmap(void		  *id,
	    struct vm_area_struct *vma)
{
	struct device_open	*o   = (struct device_open *)id;
	struct videum_device	*dev = o->dev;
	struct stream_buffer	*buf;
	int			i, n = 1;

	if (o->noncapturing)
		return -ENODEV;
	buf = mmap_stream_buffer_from_offset(dev, vma->vm_pgoff);
	if (dev->stream_contig_map)
	{/*	N buffers in one contiguous map  */
		buf = &dev->stream_buf[0];
		n = dev->stream_buffers_requested;
	}
	if (buf == NULL)
		return -EINVAL;/* no such buffer */
	if (buf->vidbuf.length * n != vma->vm_end - vma->vm_start)
		return -EINVAL;/* wrong length */
	for (i = 0; i < n; ++i)
	{
		if (!buf->requested ||
		    (buf->vidbuf.flags & V4L2_BUF_FLAG_MAPPED))
			return -EINVAL;/* not requested or already mapped */

		if (buf->vaddress != NULL)
			vfree(buf->vaddress);
		if (i == 0)
			buf->vaddress = vmalloc(buf->vidbuf.length * n);
		else
			buf->vaddress = buf[-1].vaddress + buf->vidbuf.length;
		if (buf->vaddress == NULL)
		{
			err_msg("Couldn't allocate mmap() buffer\n");
			return -ENODEV;
		}
		if (dev->dma_mode == DMAMODE_SIMPLE &&
		    !wnv_bm_build_scatter_list(dev, buf->vaddress, 
					       &buf->dma_list))
			return -ENODEV;
		buf->vidbuf.flags |= V4L2_BUF_FLAG_MAPPED;
		++dev->stream_buffers_mapped;
		++buf;
	}

	vma->vm_ops = &videum_vma_operations;
	if (vma->vm_ops->open)
		vma->vm_ops->open(vma);
	/*  Note: vma->vm_file will be set up by V4L2  */

	return 0;
}

static int
videum_poll(void	*id,
	    struct file	*file,
	    poll_table	*table)
{
	struct device_open *o = (struct device_open *)id;
	struct videum_device *dev = o->dev;

	if (o->noncapturing)
		return POLLERR;
	
	if (dev->streaming)
	{
		void	*node;
		node = v4l2_q_peek_head(&dev->stream_q_done);
		if (node != NULL)
			return (POLLIN | POLLRDNORM);/* data is ready now */
		node = v4l2_q_peek_head(&dev->stream_q_capture);
		if (node == NULL)
			return POLLERR;  /* no frames queued */
		poll_wait(file, &dev->new_video_frame, table);
		return 0;
	}

	/*  Capture is through read() call */

	if (dev->capture_completed)/* data is ready now */
		return (POLLIN | POLLRDNORM);
	capture_grab_frame(dev);/* does nothing if capture is in progress */
	if (!dev->ready_to_capture)/* Can't grab frames! */
		return POLLERR;
	poll_wait(file, &dev->new_video_frame, table);
	return 0;
}

static long
videum_read(void		*id,
	    char		*buf,
	    unsigned long	count,
	    int			noblock)
{
	struct device_open *o = (struct device_open *)id;
	struct videum_device *dev = o->dev;
	long	len		= 0;
	int	retries		= 2;
	long	my_timeout;

	if (o->noncapturing || dev->streaming)
		return -EPERM;
	capture_grab_frame(dev);/* does nothing if capture is in progress */
	if (!dev->ready_to_capture)
	{
		debug_msg("Can't grab frames!\n");
		return 0;
	}

	my_timeout = HZ / 2;
	while (len == 0)
	{
		if (noblock)
		{
			if (!dev->capture_completed)
				return -EAGAIN;
		}
		else
		{
			/* watch out for race condition going to sleep! */
			cli();
			if (!dev->capture_completed)
			{
				my_timeout = interruptible_sleep_on_timeout(
					&dev->new_video_frame, my_timeout);
			}
			sti();
		}
		if (my_timeout == 0)
		{/*	Timeout!  */
			--retries;
			if (retries == 0)
			{/*	Out of patience...  */
				debug_msg("Timeout on read. "
					  "ISR=%04X CTL=%04X\n",
					  wavi_readreg(dev, WAVI_ISR),
					  wavi_readreg(dev, WAVI_CTL));
				break;
			}
			/*  Reset and restart  */
			capture_abort(dev);
			capture_grab_frame(dev);
			continue;
		}
		len = capture_read(dev, buf, count);
	}
	//debug_msg("read %d\n", (int)len);
	return len;
}

/*
 *	Remaining initialization of video decoder etc. This is only
 *	done when the device is successfully identified and registered
 *	as a V4L2 device.
 */
static int
videum_init_done(struct v4l2_device *v)
{
	struct videum_device *dev = (struct videum_device *)v;
	int	t;
	int	i;

	/*  Initialize video input array	*/
	for (i = 0; i < VSOURCE_COUNT; ++i)
	{
		dev->source[i].input.index = i;
		dev->source[i].input.type = V4L2_INPUT_TYPE_CAMERA;
		dev->source[i].input.capability = 0;
		/*  Initialize video control properties	*/
		dev->source[i].control[VCTRL_BRIGHTNESS] =
			videum_control[VCTRL_BRIGHTNESS].default_value;
		dev->source[i].control[VCTRL_CONTRAST] =
			videum_control[VCTRL_CONTRAST].default_value;
		dev->source[i].control[VCTRL_SATURATION] =
			videum_control[VCTRL_SATURATION].default_value;
		dev->source[i].control[VCTRL_HUE] =
			videum_control[VCTRL_HUE].default_value;
	}
	strcpy(dev->source[VSOURCE_MXC].input.name, "MXC");
	strcpy(dev->source[VSOURCE_COMP].input.name, "VID/Composite");
	strcpy(dev->source[VSOURCE_SVIDEO].input.name, "SVHS/S-Video");
	strcpy(dev->source[VSOURCE_TUNER].input.name, "Tuner");
	dev->source[VSOURCE_TUNER].input.type = V4L2_INPUT_TYPE_TUNER;

	/*  Initialize the video decoder hardware	*/
	dev->videc.initialize(dev);
	t = wavi_readreg(dev, WAVI_MMA);
	if (dev->eeprom.dwHwFlags & HWFLAGS_INV_VSYNC)
		t |=  WAVI_SAA7110;
	else
		t &= ~WAVI_SAA7110;
	wavi_writereg(dev, WAVI_MMA, t);
	/*  BUG: get defaults from user somehow...  */
	dev->videc.frame_period = 333667;
	dev->videc.set_standard(dev, V4L2_STD_NTSC);
	dev->videc.set_vcrmode(dev, 0);
	set_video_input(dev, VSOURCE_MXC);

	/*  Capture parameters  */
	dev->capture.capability = V4L2_CAP_TIMEPERFRAME;
	dev->capture.capturemode = 0;
	dev->capture.extendedmode = 0;
	dev->capture.timeperframe = dev->videc.frame_period;

	/*  Default capture dimensions	*/
	dev->clientfmt.width = 160;
	dev->clientfmt.height = 120;
	dev->clientfmt.depth = 16;
	dev->clientfmt.pixelformat = V4L2_PIX_FMT_RGB565;
	dev->clientfmt.flags = 0;
	dev->clientfmt.bytesperline = 0;
	dev->clientfmt.sizeimage = 0;
	capture_new_format(dev);

	return 0;
}

/*  =====================================================================
 *	The functions below this point are only called during loading
 *	and unloading of the driver.
 */


/*
 *	D E V I C E   I N I A L I Z A T I O N   R O U T I N E S
 *
 *	These routines locate and enable the Videum, and initialize
 *	the device structure. After enabling, the WAVI register access
 *	functions and I2C functions will work.
 */

/*  Initialize v4l2_device fields	*/
static int
init_v4l2_device_fields(struct videum_device *dev)
{
	int	num	= dev - videum;

	dev->v.type = V4L2_TYPE_CAPTURE;
	dev->v.minor = unit_video[num];

	dev->v.open = videum_open;
	dev->v.close = videum_close;
	dev->v.read = videum_read;
	dev->v.write = videum_write;
	dev->v.ioctl = videum_ioctl;
	dev->v.mmap = videum_mmap;
	dev->v.poll = videum_poll;
	dev->v.initialize = videum_init_done;
	dev->v.priv = NULL;
	return 1;/* OK */
}

static void
unconfig_videum(struct videum_device *dev)
{
	capture_close(dev);
	if (dev->is_registered)
	{
		v4l2_unregister_device(&dev->v);
	}
	dev->is_registered = 0;
}

static int
config_videum(struct videum_device *dev)
{
	if (dev->wavi_type == 0)
	{
		return 0;/* No hardware for this device structure  */
	}
	init_v4l2_device_fields(dev);
	if (!find_decoder(dev))
	{
		err_msg("Bad or unrecognized video decoder\n");
		unconfig_videum(dev);
		return 0;/* failed */
	}
	if (v4l2_register_device(&dev->v) != 0)
	{
		err_msg("Couldn't register the driver.\n");
		unconfig_videum(dev);
		return 0;
	}
	dev->is_registered = 1;
	init_waitqueue_head(&dev->new_video_frame);
	return 1;
}

/*
 *	M O D U L E   I N I T   A N D   C L E A N U P
 */

int
init_module(void)
{
	int	i;

	for (i = 0; i < WNV_NBOARDS; ++i)
	{
		if (!config_videum(&videum[i]))
			break; /*  No more Videums  */
	}
	if (i == 0)
		return -ENODEV;/* cleanup will not be called */
	return 0;
}

void
cleanup_module(void)
{
	int	i;

	for (i = 0; i < WNV_NBOARDS; ++i)
		unconfig_videum(&videum[i]);
}
